// include/cfe_utils.h

#ifndef CFE_UTILS_H
#define CFE_UTILS_H

#include "utils.h"  // For StatusCode
#include <limits.h> // For LLONG_MAX
#include <stddef.h> // For size_t

// --- Structures ---

/**
 * @brief Structure to hold a CFE convergent p/q (numerator/denominator).
 */
typedef struct {
    long long p; // Numerator
    long long q; // Denominator
} Convergent;

/**
 * @brief Structure to hold a pair of integer factors.
 */
typedef struct {
    long long factor1;
    long long factor2;
} FactorPair;


// --- Constants ---

// Constants for CFE generation
#define MAX_CFE_ITERATIONS 100 // Safety limit for CFE loop iterations to prevent infinite loops for pathological inputs.
#define CFE_DBL_EPSILON 1e-15  // Tolerance for checking if CFE remainder is effectively zero.


// --- Function Prototypes ---

/**
 * @brief Calculates the Continued Fraction Expansion (CFE) convergents p_k/q_k
 *        for a given positive target value.
 *
 * @param target_value      The positive floating-point number to approximate.
 * @param max_denominator   The maximum allowed value for the denominator q_k.
 *                          Convergents with q_k > max_denominator will not be included.
 * @param convergents_out   Pointer to a Convergent* variable. On success, this
 *                          will point to a newly allocated array of convergents.
 *                          The caller is responsible for freeing this memory using free().
 *                          Set to NULL on failure.
 * @param count_out         Pointer to an int where the number of generated
 *                          convergents will be stored. Set to 0 on failure.
 *
 * @return StatusCode indicating the result:
 *         - SUCCESS: Convergents calculated successfully. Note: This may be returned
 *                    even if an internal overflow occurred during calculation, provided
 *                    at least one valid convergent was found before the overflow; the
 *                    returned list will be partial in that case.
 *         - ERR_INVALID_INPUT: target_value <= 0, is non-finite, or output pointers are NULL.
 *         - ERR_MEMORY_ALLOC: Memory allocation for the convergents array failed.
 *         - ERR_OVERFLOW: Arithmetic overflow occurred during calculation (e.g., p_k or q_k exceeded LLONG_MAX).
 *                         If count_out > 0, see SUCCESS note above.
 *         - ERR_CALCULATION: Other calculation issues, or no convergents found within limits.
 */
StatusCode calculate_cfe_convergents(double target_value, long long max_denominator,
                                   Convergent** convergents_out, int* count_out);


/**
 * @brief Finds all unique pairs of positive integer factors (f1, f2) for a
 *        given positive number, such that f1 * f2 = number.
 *        Includes the pair (1, number). Iterates up to sqrt(number).
 *
 * @param number            The positive integer to factor (must be > 0).
 * @param factor_pairs_out  Pointer to a FactorPair* variable. On success, this
 *                          will point to a newly allocated array of factor pairs.
 *                          The caller is responsible for freeing this memory using free().
 *                          Set to NULL on failure.
 * @param count_out         Pointer to an int where the number of found factor pairs
 *                          will be stored. Set to 0 on failure.
 *
 * @return StatusCode indicating the result:
 *         - SUCCESS: Factorization successful.
 *         - ERR_INVALID_INPUT: number <= 0 or output pointers are NULL.
 *         - ERR_MEMORY_ALLOC: Memory allocation for the factor pairs array failed.
 *         - ERR_OVERFLOW: Internal capacity calculations overflowed size_t.
 *         - ERR_CALCULATION: Should not generally occur for valid inputs.
 */
StatusCode find_integer_factor_pairs(long long number,
                                     FactorPair** factor_pairs_out, int* count_out);


#endif // CFE_UTILS_H
