// include/input.h

#ifndef INPUT_H
#define INPUT_H

#include <stdbool.h> // For bool return type
#include "utils.h"   // For StatusCode (needed in prototypes)
#include "resampling_common.h" // For HARD_MAX_LM_LIMIT, LARGE_LM_THRESHOLD etc.

// --- Constants ---
#define MAX_INPUT_BUFFER_SIZE 256 // Max characters for line input buffer (including null terminator)

// --- Enums ---

/**
 * @brief Specifies whether to use one or two resampler stages.
 */
typedef enum {
    ONE_RESAMPLER = 1,
    TWO_RESAMPLERS = 2
} ResamplerMode;

/**
 * @brief Specifies the search strategy for the two-resampler mode.
 */
typedef enum {
    SEARCH_DEFAULT,       /**< Use user-provided max L/M per stage, find best approximation or first match. */
    SEARCH_MINIMAL_SUM    /**< Use internal hard limit, find solution with absolute minimal L1+M1+L2+M2 sum. */
} SearchMode;

/**
 * @brief Defines how to prioritize targets when an intermediate rate is specified
 *        in the two-resampler mode.
 */
typedef enum {
    PRIORITIZE_STRICT_INTERMEDIATE = 1, /**< Must meet intermediate rate first, then optimize final rate. */
    PRIORITIZE_FINAL_RATE = 2,          /**< Must meet final rate first, then choose closest intermediate. */
    PRIORITIZE_BALANCED_DUAL = 3,       /**< Try for both exact matches, fallback to final rate match. */
    PRIORITIZE_BEST_OVERALL = 4         /**< Ignore intermediate rate for selection, use overall search mode logic. */
} IntermediatePriority;


// --- Structures ---

/**
 * @brief Holds all configuration parameters gathered from the user.
 */
typedef struct {
    // Core parameters
    double input_rate;                 /**< Input sample rate (Hz). */
    double output_rate;                /**< Target output sample rate (Hz). */
    ResamplerMode num_resamplers;      /**< ONE_RESAMPLER or TWO_RESAMPLERS. */

    // Two-resampler specific configuration
    SearchMode search_mode;            /**< SEARCH_DEFAULT or SEARCH_MINIMAL_SUM. */
    bool allow_m1_greater_than_l1;     /**< Flag to allow M1 > L1 in the first stage. */
    long long max_lm_per_stage;        /**< Max L/M per stage (user input for SEARCH_DEFAULT, else hard limit). */

    // Intermediate rate targeting (two-resampler mode)
    double target_intermediate_rate;   /**< Optional target rate after stage 1 (Hz). */
    bool intermediate_rate_targeted;   /**< True if user provided a target intermediate rate. */
    IntermediatePriority intermediate_priority; /**< How to prioritize intermediate vs final rate targets. */

} UserInputs;

// --- Function Prototypes ---

/**
 * @brief Prompts the user and reads a positive double value from standard input.
 *        Handles errors, retries on invalid input, and checks for shutdown requests.
 *
 * @param prompt The message to display to the user.
 * @param value  Pointer to a double where the validated input will be stored.
 * @return SUCCESS, FAILURE (on shutdown), or ERR_INVALID_INPUT (on persistent errors/EOF).
 */
StatusCode get_positive_double(const char* prompt, double* value);

/**
 * @brief Prompts the user and reads an integer that must be one of the allowed values.
 *        Handles errors, retries, and checks for shutdown requests.
 *
 * @param prompt          The message to display to the user.
 * @param allowed_values  An array of integers that are considered valid input.
 * @param num_allowed     The number of elements in the allowed_values array.
 * @param value           Pointer to an int where the validated input will be stored.
 * @return SUCCESS, FAILURE (on shutdown), or ERR_INVALID_INPUT (on persistent errors/EOF).
 */
StatusCode get_specific_integer(const char* prompt, const int allowed_values[], int num_allowed, int* value);

/**
 * @brief Prompts the user and reads a positive long long integer value from standard input.
 *        Handles errors, retries, and checks for shutdown requests.
 *
 * @param prompt The message to display to the user.
 * @param value  Pointer to a long long where the validated input will be stored.
 * @return SUCCESS, FAILURE (on shutdown), or ERR_INVALID_INPUT (on persistent errors/EOF).
 */
StatusCode get_positive_long_long(const char* prompt, long long* value);

/**
 * @brief Gathers all necessary user inputs interactively and populates the UserInputs struct.
 *        Calls the specific input helper functions. Handles shutdown requests.
 *
 * @param inputs Pointer to the UserInputs struct to be filled.
 * @return SUCCESS if all inputs gathered successfully, FAILURE (on shutdown), or an error code
 *         propagated from the input helper functions.
 */
StatusCode get_user_inputs(UserInputs* inputs);

/**
 * @brief Prompts the user for a yes/no confirmation.
 *        Accepts 'y'/'Y' for yes, 'n'/'N' or Enter (empty input) for no.
 *        Handles errors, retries, and checks for shutdown requests.
 *
 * @param prompt The confirmation question to display.
 * @return true if the user confirmed ('y'/'Y'), false otherwise ('n'/'N'/Enter/error/shutdown).
 */
bool get_confirmation(const char* prompt);


#endif // INPUT_H
