// include/resampling_common.h

#ifndef RESAMPLING_COMMON_H
#define RESAMPLING_COMMON_H

#include <stdbool.h> // For bool in TwoResamplerResult
#include <stdint.h>  // For SIZE_MAX (used indirectly via checks in CFE utils)
#include <math.h>    // For NAN, INFINITY initialization
#include <limits.h>  // For LLONG_MAX (needed for sum_coeffs init)

// --- Core Constants ---
#define MATCH_TOLERANCE 1e-9        // Relative tolerance for considering overall ratio match successful
#define INTERMEDIATE_MATCH_TOLERANCE 1e-9 // Relative tolerance for intermediate rate match (can be adjusted)
#define SINGLE_RESAMPLER_TOLERANCE_FACTOR 10.0 // Looser tolerance multiplier for single resampler "exactness" check vs target
#define ARBITRARY_SUGGEST_ERROR_THRESHOLD (MATCH_TOLERANCE * 1000.0) // Threshold for suggesting arbitrary due to poor approximation error (e.g., 1e-6)

// --- Limit Constants ---
#define LARGE_LM_THRESHOLD 1000     // Threshold for warning about large L/M calculation/filter cost during input confirmation
#define SUGGEST_ARBITRARY_LM_THRESHOLD 1000 // Threshold for suggesting arbitrary resampling in output due to large L/M factors themselves
#define HARD_MAX_LM_LIMIT 65535     // Practical upper limit for L/M
#define MAX_REASONABLE_RATE 1.0e13  // Max input rate check
#define MIN_REASONABLE_RATE 1.0     // Min input rate check


// --- Structures ---

/**
 * @brief Holds the best result found by the two-resampler CFE+Factor search.
 */
typedef struct {
    long long best_l1;
    long long best_m1;
    long long best_l2;
    long long best_m2;
    double best_combined_ratio;
    double min_error;               // Absolute overall ratio error
    bool match_found;               // Overall match found within tolerance
    bool result_found;              // Any valid result found
    long long sum_coeffs;           // Sum L1+M1+L2+M2, initialized high (LLONG_MAX)
    // Intermediate Rate Results
    double achieved_intermediate_rate;
    double intermediate_rate_error;    // Absolute intermediate rate error vs target
    bool intermediate_match_found;   // Intermediate match found within tolerance
} TwoResamplerResult;


#endif // RESAMPLING_COMMON_H
