// include/single_resampler.h

#ifndef SINGLE_RESAMPLER_H
#define SINGLE_RESAMPLER_H

#include "input.h" // For UserInputs struct
#include "utils.h" // For StatusCode and max_long_long helper
#include "resampling_common.h" // For constants like HARD_MAX_LM_LIMIT

// --- Function Prototypes ---

/**
 * @brief Calculates the best rational approximation L/M for a single-stage
 *        resampler using Continued Fraction Expansion (CFE).
 *
 * Uses CFE up to a maximum denominator M determined by HARD_MAX_LM_LIMIT.
 * Checks the global shutdown flag during execution.
 *
 * @param input_rate          The input sample rate (Hz, must be positive finite).
 * @param output_rate         The target output sample rate (Hz, must be positive finite).
 * @param l_out               Pointer to store the calculated interpolation factor L.
 * @param m_out               Pointer to store the calculated decimation factor M.
 * @param achieved_output_rate Pointer to store the actual output rate achievable with L/M.
 *
 * @return SUCCESS if calculation is successful,
 *         FAILURE if shutdown was requested during calculation,
 *         ERR_INVALID_INPUT if rates are invalid or output pointers are NULL,
 *         ERR_MEMORY_ALLOC if CFE memory allocation fails,
 *         ERR_CALCULATION on CFE failure, non-positive L/M, or other calculation issues,
 *         ERR_OVERFLOW if CFE internal calculations overflow.
 */
StatusCode calculate_single_resampler(double input_rate, double output_rate, long long* l_out, long long* m_out, double* achieved_output_rate);

/**
 * @brief Prints the results of the single-stage resampler calculation in a formatted way.
 *
 * Includes target/achieved rates, L/M values, ratio, errors (absolute, relative, PPM).
 * Adds cautionary notes for significant approximation errors or large L/M values.
 * Checks the global shutdown flag.
 * **User-facing output format remains unchanged.**
 *
 * @param input_rate         The original input sample rate.
 * @param output_rate        The original target output sample rate.
 * @param l                  The calculated interpolation factor L (must be > 0).
 * @param m                  The calculated decimation factor M (must be > 0).
 * @param achieved_output_rate The actual output rate achieved with L/M.
 */
void print_single_resampler_result(double input_rate, double output_rate, long long l, long long m, double achieved_output_rate);

/**
 * @brief Handles the entire process for the single resampler mode.
 *
 * Prints context, calls the calculation function, and prints the results (or failure message).
 * Checks the global shutdown flag.
 *
 * @param inputs Pointer to the UserInputs struct containing rates.
 * @return SUCCESS if calculation and printing were successful,
 *         FAILURE if shutdown was requested,
 *         or an error code propagated from calculate_single_resampler.
 */
StatusCode handle_one_resampler(const UserInputs* inputs);


#endif // SINGLE_RESAMPLER_H
