// include/utils.h

#ifndef UTILS_H
#define UTILS_H

#include <stdio.h>  // For vfprintf, stderr in print_stderr
#include <stdlib.h> // For llabs (though helper might handle directly)
#include <stdarg.h> // For va_list
#include <limits.h> // For INT_MAX etc.

// --- Enums ---

/**
 * @brief Status codes used across modules to indicate operation results.
 */
typedef enum {
    SUCCESS = 0,           /**< Operation completed successfully. */
    FAILURE = -1,          /**< General failure, not found, user cancellation, or shutdown. */
    ERR_MEMORY_ALLOC = -2, /**< Memory allocation failed (e.g., malloc, realloc). */
    ERR_INVALID_INPUT = -3,/**< Input data provided was invalid or unsuitable. */
    ERR_CALCULATION = -6,  /**< Internal calculation error occurred (e.g., division by zero, non-convergence). */
    ERR_FILE_IO = -11,     /**< Error during file input/output (not currently used). */
    ERR_OVERFLOW = -12     /**< Arithmetic overflow detected during calculation. */
} StatusCode;


// --- Function Prototypes ---

/**
 * @brief Prints a formatted message to stderr, preceded and followed by a newline.
 *        Useful for consistent error/warning message formatting.
 *
 * @param format The format string (like printf).
 * @param ...    Variable arguments for the format string.
 */
void print_stderr(const char* format, ...);

/**
 * @brief Converts a StatusCode enum value to its string representation.
 *
 * @param code The StatusCode value.
 * @return A constant string describing the status code. Returns "Unknown Error Code"
 *         for values not explicitly handled.
 */
const char* status_code_to_string(StatusCode code);

/**
 * @brief Returns the maximum of two long long values.
 *        Handles negative inputs defensively (returns the larger, or b if a is neg, or -1 if both neg).
 *        Needed because standard max() often works on int or double.
 *
 * @param a The first long long value.
 * @param b The second long long value.
 * @return The larger of a and b, or a suitable value if one/both are negative.
 */
long long max_long_long(long long a, long long b);


#endif // UTILS_H
