// src/cfe_utils.c

#include "cfe_utils.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdbool.h>
#include <stdint.h> // For SIZE_MAX, ULLONG_MAX
#include <limits.h> // For LLONG_MAX

// --- Constants for Array Resizing ---
#define INITIAL_CONVERGENT_CAPACITY 16 // Sensible starting size for convergents
#define INITIAL_FACTOR_PAIR_CAPACITY 8 // Modest starting size for factor pairs

// --- Static Helper Prototypes ---

// Dynamic array management for Convergent structs
static StatusCode resize_convergent_array(Convergent** convergents_ptr, size_t* capacity_ptr, int current_count);
static StatusCode add_convergent(Convergent** convergents_ptr, size_t* capacity_ptr, int* count_ptr, long long p, long long q);

// Dynamic array management for FactorPair structs
static StatusCode resize_factor_pair_array(FactorPair** factors_ptr, size_t* capacity_ptr, int current_count);
static StatusCode add_factor_pair(FactorPair** factors_ptr, size_t* capacity_ptr, int* count_ptr, long long f1, long long f2);


// --- Public Function Implementation: CFE ---

/**
 * @brief Calculates Continued Fraction Expansion convergents.
 * @details Implements the standard CFE algorithm, generating p_k/q_k approximations.
 *          Handles potential overflows and limits by max_denominator.
 *          Dynamically allocates and resizes the output array.
 *          See header file for detailed parameter and return code descriptions.
 */
StatusCode calculate_cfe_convergents(double target_value, long long max_denominator,
                                   Convergent** convergents_out, int* count_out)
{
    // --- Input Validation ---
    if (!convergents_out || !count_out) return ERR_INVALID_INPUT;
    *convergents_out = NULL;
    *count_out = 0;

    if (!isfinite(target_value) || target_value <= 0.0) {
        return ERR_INVALID_INPUT; // Target must be positive and finite
    }
    if (max_denominator <= 0) {
        // Allow calculation even if max_denominator is non-positive,
        // but no convergents beyond q=0 or q=1 will be generated.
        // Let's enforce positive for clarity, as q=0 is never valid.
        return ERR_INVALID_INPUT;
    }

    // --- Initialization ---
    StatusCode status = SUCCESS;
    Convergent* convergents = NULL;
    int count = 0;
    size_t capacity = 0;

    // CFE algorithm state variables (Knuth notation)
    long long p_prev2 = 0, q_prev2 = 1; // p_{-2}, q_{-2}
    long long p_prev1 = 1, q_prev1 = 0; // p_{-1}, q_{-1}
    long long pk = 0, qk = 0;           // p_k, q_k (current convergent)
    double remainder = target_value;    // Current value being expanded (alpha_k)
    long long ak_ll = 0;                // Integer part of remainder (a_k)

    // Allocate initial dynamic array
    status = resize_convergent_array(&convergents, &capacity, 0);
    if (status != SUCCESS) {
        goto cleanup_cfe; // Allocation failed
    }

    // --- CFE Iteration Loop ---
    for (int k = 0; k < MAX_CFE_ITERATIONS; ++k) {
        // Check for termination conditions
        if (fabs(remainder) < CFE_DBL_EPSILON) {
            break; // Remainder is effectively zero, exact fraction found
        }
        if (!isfinite(remainder) || remainder <= 0) {
             // Should not happen with valid initial input, but check defensively
             status = ERR_CALCULATION;
             goto cleanup_cfe;
        }

        // Calculate integer part a_k
        double ak_dbl = floor(remainder);
        if (ak_dbl > (double)LLONG_MAX || ak_dbl < 0) { // Check range before casting
             status = ERR_OVERFLOW; goto cleanup_cfe; // a_k must fit in positive long long
        }
        ak_ll = (long long)ak_dbl;
        if (ak_ll < 0) { // Should be caught above, but double check
            status = ERR_CALCULATION; goto cleanup_cfe;
        }

        // Calculate next convergent (pk, qk) using recurrence relations
        // p_k = a_k * p_{k-1} + p_{k-2}
        // q_k = a_k * q_{k-1} + q_{k-2}
        // Check for overflow BEFORE multiplication/addition

        // Check overflow for pk = ak_ll * p_prev1 + p_prev2
        // Condition: ak_ll > (LLONG_MAX - p_prev2) / p_prev1 (rearranged to avoid overflow during check)
        // Ensure p_prev1 > 0 to avoid division by zero (p_prev1 starts at 1 and grows)
        if (p_prev1 > 0 && ak_ll > 0 && p_prev1 > (LLONG_MAX - p_prev2) / ak_ll) {
            status = ERR_OVERFLOW;
            break; // Stop iteration, potential partial results may be returned
        }
        pk = ak_ll * p_prev1 + p_prev2;

        // Check overflow for qk = ak_ll * q_prev1 + q_prev2
        // Condition: ak_ll > (LLONG_MAX - q_prev2) / q_prev1
        // Ensure q_prev1 >= 0 (starts at 0, grows non-negatively)
        // Special case: if q_prev1 is 0, qk = q_prev2 (which is 1 initially), no overflow risk there.
        if (q_prev1 > 0 && ak_ll > 0 && q_prev1 > (LLONG_MAX - q_prev2) / ak_ll) {
             status = ERR_OVERFLOW;
             break; // Stop iteration
        }
        qk = ak_ll * q_prev1 + q_prev2;

        // Check for potential negative qk (should only happen on unexpected overflow)
        if (qk < 0) {
            status = ERR_OVERFLOW;
            break; // Should be caught by checks above, but safeguard
        }

        // Check against max_denominator constraint *before* adding
        if (qk > max_denominator) {
             break; // Denominator limit exceeded, stop generating
        }
        // Also handle qk=0 case, which is invalid for a convergent denominator
        if (qk == 0 && k > 0) { // q0 can be 0 (from q-1=0), but subsequent qk must be > 0
             status = ERR_CALCULATION; // Should not happen with positive target
             goto cleanup_cfe;
        }


        // Add the new convergent (pk, qk) to the dynamic array
        // Skip adding the initial p0/q0 = a0*1+0 / a0*0+1 = a0/1 if qk=0 (which should only be q_prev1=0 case)
        // The first meaningful convergent is p1/q1
        // Let's add p0/q0 = floor(target)/1 as well, if qk=1 initially
        if (qk > 0) { // Ensure valid denominator before adding
            StatusCode add_status = add_convergent(&convergents, &capacity, &count, pk, qk);
            if (add_status != SUCCESS) {
                status = add_status; // Propagate memory/overflow error from add
                goto cleanup_cfe;
            }
        } else if (k > 0) { // qk = 0 is invalid after first step
             status = ERR_CALCULATION;
             goto cleanup_cfe;
        }


        // Update remainder for next iteration: remainder = 1.0 / (remainder - ak_dbl)
        remainder = remainder - ak_dbl; // Fractional part
        if (fabs(remainder) < CFE_DBL_EPSILON) {
            break; // Remainder is zero, we are done
        }
        if (remainder <= 0) { // Defensive check
             status = ERR_CALCULATION; goto cleanup_cfe;
        }
        remainder = 1.0 / remainder; // Take reciprocal
        if (!isfinite(remainder)) {
            // Could happen if fractional part was extremely small, leading to large reciprocal
            break; // Treat as termination
        }

        // Update history for next iteration
        p_prev2 = p_prev1; q_prev2 = q_prev1;
        p_prev1 = pk;      q_prev1 = qk;

    } // End CFE Iteration Loop

    // --- Post-Loop Processing ---

    // Handle overflow termination specifically:
    // If overflow occurred but we have some results, return them as SUCCESS.
    // If overflow occurred with no results, it's a fatal error.
    if (status == ERR_OVERFLOW && count > 0) {
        status = SUCCESS; // Return the valid convergents found so far
    } else if (status == ERR_OVERFLOW && count == 0) {
        goto cleanup_cfe; // Overflow on very first steps is fatal for output
    }
    // Handle normal termination or other non-fatal errors (like MAX_ITERATIONS reached)

    // --- Final Output ---
    if (status == SUCCESS && count > 0) {
        // Success path: Transfer ownership of the allocated array
        *convergents_out = convergents;
        *count_out = count;
        convergents = NULL; // Prevent free in cleanup
    } else if (status == SUCCESS && count == 0) {
        // No convergents found meeting criteria (e.g., max_denominator too small)
        // This is not necessarily an error, but might be considered a calculation "failure"
        // in the context of needing a result. Let's map it to ERR_CALCULATION.
        status = ERR_CALCULATION;
        goto cleanup_cfe;
    } else {
         // Any other error path (memory, calculation, fatal overflow)
         goto cleanup_cfe;
    }

    // Fallthrough success return is here
    return SUCCESS;

cleanup_cfe:
    // Free memory only if it wasn't transferred
    if (convergents != NULL) {
        free(convergents);
    }
    // Ensure output pointers are NULL on any failure path
    *convergents_out = NULL;
    *count_out = 0;
    // Return the determined error status
    return status;
}


// --- Public Function Implementation: Factorization ---

/**
 * @brief Finds all unique pairs of positive integer factors (f1, f2) for a number.
 * @details Uses trial division up to sqrt(number). Efficient for moderately sized numbers.
 *          Dynamically allocates and resizes the output array.
 *          Note: Performance degrades for very large numbers, especially semi-primes (O(sqrt N)).
 *          See header file for detailed parameter and return code descriptions.
 */
StatusCode find_integer_factor_pairs(long long number,
                                     FactorPair** factor_pairs_out, int* count_out)
{
    // --- Input Validation ---
    if (!factor_pairs_out || !count_out) return ERR_INVALID_INPUT;
    *factor_pairs_out = NULL;
    *count_out = 0;

    if (number <= 0) {
        // fprintf(stderr, "Error (Factor): Number to factor must be positive.\n"); // Keep prompts out
        return ERR_INVALID_INPUT;
    }

    // --- Initialization ---
    StatusCode status = SUCCESS;
    FactorPair* factors = NULL;
    int count = 0;
    size_t capacity = 0;

    // Allocate initial dynamic array
    status = resize_factor_pair_array(&factors, &capacity, 0);
    if (status != SUCCESS) {
        goto cleanup_factor;
    }

    // --- Factorization Loop ---
    // Calculate loop limit carefully to avoid overflow and precision issues.
    // Note: O(sqrt(N)) complexity. Can be slow for large N.
    double sqrt_num_dbl = sqrt((double)number);
    long long limit = (long long)sqrt_num_dbl;

    // Correction for potential floating-point inaccuracies near integer boundaries
    // Check if (limit + 1)^2 <= number without overflow.
    if (limit >= 0 && limit < LLONG_MAX) { // Ensure limit+1 doesn't overflow itself
        long long limit_plus_1 = limit + 1;
        // Check potential multiplication overflow: limit_plus_1 <= LLONG_MAX / limit_plus_1
        // Ensure limit_plus_1 > 0 to avoid division by zero (true if limit >= 0)
        if (limit_plus_1 > 0 && limit_plus_1 <= LLONG_MAX / limit_plus_1) {
            // Now safe to multiply
            if (limit_plus_1 * limit_plus_1 <= number) {
                limit++; // Increment limit if (limit+1)^2 is representable and <= number
            }
        }
        // If limit_plus_1 * limit_plus_1 would overflow, number must be very large.
        // The original sqrt approximation is likely sufficient, no need to increment.
    }

    // Final checks on the calculated limit
    if (limit <= 0 && number > 0) limit = 1; // Handle sqrt(1) or small number cases
    if (limit > number) limit = number; // Ensure limit doesn't exceed number itself

    // Iterate through potential factors f1 up to the calculated limit
    for (long long f1 = 1; f1 <= limit; ++f1) {
        if (number % f1 == 0) {
            // f1 is a factor, find the corresponding factor f2
            long long f2 = number / f1;
            // Add the pair (f1, f2) to the results array
            StatusCode add_status = add_factor_pair(&factors, &capacity, &count, f1, f2);
            if (add_status != SUCCESS) {
                status = add_status; // Propagate memory/overflow error
                goto cleanup_factor;
            }
            // No need to add (f2, f1) separately, as the loop only goes up to sqrt(number).
            // If f1 == f2 (perfect square), it's added once correctly.
        }
    }

    // --- Final Output ---
     if (status == SUCCESS && count > 0) {
         // Success path: Transfer ownership of the allocated array
        *factor_pairs_out = factors;
        *count_out = count;
        factors = NULL; // Prevent free in cleanup
     } else if (status == SUCCESS && count == 0) {
         // This should only happen if number=1, where loop limit is 1 and f1=1 gives f2=1.
         // If it happens otherwise, it's an error.
         if (number == 1 && count == 1) { // Check if number=1 case worked correctly
             *factor_pairs_out = factors; *count_out = count; factors = NULL;
         } else {
             status = ERR_CALCULATION; // Should not happen for number > 1
             goto cleanup_factor;
         }
     } else {
         // Any other error path (memory allocation, etc.)
         goto cleanup_factor;
     }

    // Fallthrough success return is here
    return SUCCESS;

cleanup_factor:
    // Free memory only if it wasn't transferred
    if (factors != NULL) {
        free(factors);
    }
    // Ensure output pointers are NULL on any failure path
    *factor_pairs_out = NULL;
    *count_out = 0;
    // Return the determined error status
    return status;
}


// --- Static Helper Implementation: Resize Convergent Array ---
/**
 * @brief Resizes the dynamic array for Convergent structs if needed.
 *        Doubles capacity or sets initial capacity. Handles potential overflows.
 */
static StatusCode resize_convergent_array(Convergent** convergents_ptr, size_t* capacity_ptr, int current_count) {
     if (!convergents_ptr || !capacity_ptr) return ERR_INVALID_INPUT; // Should not happen internally

     // Check if resize is needed (using size_t for comparison)
     if ((size_t)current_count < *capacity_ptr) {
         return SUCCESS; // Enough space already
     }

     // Calculate new capacity (double or initial)
     size_t new_capacity = (*capacity_ptr == 0) ? INITIAL_CONVERGENT_CAPACITY : (*capacity_ptr * 2);

     // Check for arithmetic overflow when doubling capacity
     if (new_capacity <= *capacity_ptr && *capacity_ptr > 0) { // Check if doubling overflowed or didn't increase
         return ERR_OVERFLOW;
     }
     // Check if requested size exceeds memory limits (SIZE_MAX / element_size)
     if (new_capacity > SIZE_MAX / sizeof(Convergent)) {
         return ERR_OVERFLOW; // Requested allocation size is too large
     }

     // Attempt reallocation
     Convergent* temp = realloc(*convergents_ptr, new_capacity * sizeof(Convergent));
     if (!temp) {
         // If realloc fails, the original pointer *convergents_ptr is still valid
         return ERR_MEMORY_ALLOC;
     }

     // Update pointer and capacity on success
     *convergents_ptr = temp;
     *capacity_ptr = new_capacity;
     return SUCCESS;
}

// --- Static Helper Implementation: Add Convergent ---
/**
 * @brief Adds a convergent (p, q) to the dynamic array, resizing if necessary.
 */
static StatusCode add_convergent(Convergent** convergents_ptr, size_t* capacity_ptr, int* count_ptr, long long p, long long q) {
    // Ensure array has space, resizing if needed
    StatusCode status = resize_convergent_array(convergents_ptr, capacity_ptr, *count_ptr);
    if (status != SUCCESS) {
        return status; // Propagate error (memory/overflow)
    }

    // Add the new convergent data
    (*convergents_ptr)[*count_ptr].p = p;
    (*convergents_ptr)[*count_ptr].q = q;
    (*count_ptr)++; // Increment count *after* successfully adding

    return SUCCESS;
}

// --- Static Helper Implementation: Resize Factor Pair Array ---
/**
 * @brief Resizes the dynamic array for FactorPair structs if needed.
 *        Doubles capacity or sets initial capacity. Handles potential overflows.
 */
static StatusCode resize_factor_pair_array(FactorPair** factors_ptr, size_t* capacity_ptr, int current_count) {
     if (!factors_ptr || !capacity_ptr) return ERR_INVALID_INPUT; // Internal error check

     // Check if resize is needed
     if ((size_t)current_count < *capacity_ptr) {
         return SUCCESS; // Enough space
     }

     // Calculate new capacity
     size_t new_capacity = (*capacity_ptr == 0) ? INITIAL_FACTOR_PAIR_CAPACITY : (*capacity_ptr * 2);

     // Check for arithmetic overflow
     if (new_capacity <= *capacity_ptr && *capacity_ptr > 0) {
         return ERR_OVERFLOW;
     }
     // Check for allocation size overflow
     if (new_capacity > SIZE_MAX / sizeof(FactorPair)) {
         return ERR_OVERFLOW;
     }

     // Attempt reallocation
     FactorPair* temp = realloc(*factors_ptr, new_capacity * sizeof(FactorPair));
     if (!temp) {
         return ERR_MEMORY_ALLOC;
     }

     // Update pointer and capacity
     *factors_ptr = temp;
     *capacity_ptr = new_capacity;
     return SUCCESS;
}

// --- Static Helper Implementation: Add Factor Pair ---
/**
 * @brief Adds a factor pair (f1, f2) to the dynamic array, resizing if necessary.
 */
static StatusCode add_factor_pair(FactorPair** factors_ptr, size_t* capacity_ptr, int* count_ptr, long long f1, long long f2) {
    // Ensure array has space
    StatusCode status = resize_factor_pair_array(factors_ptr, capacity_ptr, *count_ptr);
    if (status != SUCCESS) {
        return status; // Propagate error
    }

    // Add the new factor pair data
    (*factors_ptr)[*count_ptr].factor1 = f1;
    (*factors_ptr)[*count_ptr].factor2 = f2;
    (*count_ptr)++; // Increment count

    return SUCCESS;
}
