// src/input.c

#include "input.h"
#include "signal_handler.h"
#include "utils.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>

#ifdef _WIN32
#include <windows.h>
#include <conio.h> // For _putch, useful for echo
#endif

// --- Static Helper Prototypes ---
static void clear_stdin_buffer(void);
static const char* search_mode_to_string(SearchMode mode);

// --- Platform-Specific Input Reading Function Prototypes ---
#ifdef _WIN32
// static StatusCode read_line_windows(const char* prompt, char* buffer, size_t buffer_size); // REMOVED OLD ONE
static StatusCode read_line_windows_raw(const char* prompt, char* buffer, size_t buffer_size); // NEW ONE
#else
static StatusCode read_line_fgets(const char* prompt, char* buffer, size_t buffer_size);
#endif


// --- Static Helper Implementation ---
// ... (clear_stdin_buffer, search_mode_to_string - unchanged) ...
static void clear_stdin_buffer(void) {
    // Note: This function reads from stdin, which we are bypassing in read_line_windows_raw.
    // It might not be necessary or effective after using ReadConsoleInput unless
    // there's a mix of input methods, but leave it for now.
    int c;
    // We should technically use console functions to clear if bypassing stdin fully,
    // but let's stick to the plan for now.
    while ((c = getchar()) != '\n' && c != EOF) {
        if (g_shutdown_requested) break;
    }
     // Flush any remaining console input events if using raw mode? More complex.
     // For now, assume this is sufficient for clearing *after* an error.
}

static const char* search_mode_to_string(SearchMode mode) {
    switch (mode) {
        case SEARCH_DEFAULT:     return "custom limit criteria";
        case SEARCH_MINIMAL_SUM: return "lowest sum";
        default:                 return "unknown mode";
    }
}


// --- Input Reading Implementation ---

#ifdef _WIN32

/**
 * @brief Reads a line using low-level ReadConsoleInput, handling basic
 *        editing (backspace) and echo. Interruptible via shutdown event.
 */
static StatusCode read_line_windows_raw(const char* prompt, char* buffer, size_t buffer_size) {
    HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
    HANDLE hStdout = GetStdHandle(STD_OUTPUT_HANDLE); // Needed for echo/backspace
    HANDLE handles[2];
    DWORD initial_mode = 0;
    DWORD new_mode = 0;
    DWORD chars_read;
    INPUT_RECORD ir_in_buf[128]; // Read multiple records at once
    size_t buffer_pos = 0;
    StatusCode final_status = ERR_CALCULATION; // Default to error

    if (hStdin == INVALID_HANDLE_VALUE || hStdout == INVALID_HANDLE_VALUE || buffer_size == 0) {
        print_stderr("Failed to get standard handles.");
        return ERR_CALCULATION;
    }
    if (g_hShutdownEvent == NULL) {
        print_stderr("Internal Error: Shutdown event handle is NULL.");
        return ERR_CALCULATION;
    }

    // --- Save initial console mode and set raw mode ---
    if (!GetConsoleMode(hStdin, &initial_mode)) {
        print_stderr("Failed to get console mode (Error %lu).", GetLastError());
        return ERR_CALCULATION;
    }
    new_mode = initial_mode;
    // Disable line input, echo input. Enable processed input (for Ctrl+C). Window/Mouse not needed.
    new_mode &= ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
    new_mode |= ENABLE_PROCESSED_INPUT; // Ensure Ctrl+C is processed by system
    // Optional: Consider ENABLE_WINDOW_INPUT / ENABLE_MOUSE_INPUT if needed later, disable now.
    new_mode &= ~(ENABLE_WINDOW_INPUT | ENABLE_MOUSE_INPUT);

    if (!SetConsoleMode(hStdin, new_mode)) {
        print_stderr("Failed to set raw console mode (Error %lu).", GetLastError());
        // Don't restore mode here, as we failed to set it.
        return ERR_CALCULATION;
    }

    // --- Ensure buffer is clear initially ---
    buffer[0] = '\0';

    // --- Print Prompt ---
    if (prompt) {
        DWORD written;
        WriteConsoleA(hStdout, prompt, (DWORD)strlen(prompt), &written, NULL);
    }

    handles[0] = hStdin;
    handles[1] = g_hShutdownEvent;

    // --- Input Loop ---
    while (buffer_pos < buffer_size - 1) { // Leave space for null terminator
        // Check flag before waiting
        if (g_shutdown_requested) {
            final_status = FAILURE;
            goto cleanup;
        }

        DWORD waitResult = WaitForMultipleObjects(2, handles, FALSE, INFINITE);

        // Check flag immediately after waiting
        if (g_shutdown_requested) {
            final_status = FAILURE;
            goto cleanup;
        }

        if (waitResult == WAIT_OBJECT_0) { // Input available
            if (!ReadConsoleInput(hStdin, ir_in_buf, 128, &chars_read)) {
                 print_stderr("ReadConsoleInput failed (Error %lu).", GetLastError());
                 final_status = ERR_CALCULATION;
                 goto cleanup;
            }

             // Check flag after potentially blocking ReadConsoleInput
             if (g_shutdown_requested) {
                  final_status = FAILURE;
                  goto cleanup;
             }

            for (DWORD i = 0; i < chars_read; ++i) {
                if (ir_in_buf[i].EventType == KEY_EVENT && ir_in_buf[i].Event.KeyEvent.bKeyDown) {
                    KEY_EVENT_RECORD ker = ir_in_buf[i].Event.KeyEvent;
                    char ascii_char = ker.uChar.AsciiChar; // Use ASCII for simplicity here
                    WORD vk_code = ker.wVirtualKeyCode;
                    DWORD written;

                    if (vk_code == VK_RETURN) { // Enter key
                        // Echo newline
                        WriteConsoleA(hStdout, "\r\n", 2, &written, NULL);
                        buffer[buffer_pos] = '\0'; // Null terminate
                        final_status = SUCCESS;
                        goto cleanup; // Exit loop and function
                    }
                    else if (vk_code == VK_BACK) { // Backspace key
                        if (buffer_pos > 0) {
                            buffer_pos--;
                            buffer[buffer_pos] = '\0'; // Erase from buffer
                            // Erase character visually from console
                            WriteConsoleA(hStdout, "\b \b", 3, &written, NULL);
                        }
                    }
                    // Check for printable ASCII character (basic check)
                    else if (ascii_char >= ' ' && ascii_char <= '~') {
                        // Check if buffer has space (including null terminator)
                        if (buffer_pos < buffer_size - 1) {
                             buffer[buffer_pos++] = ascii_char;
                             buffer[buffer_pos] = '\0'; // Keep null terminated
                             // Echo character visually
                             WriteConsoleA(hStdout, &ascii_char, 1, &written, NULL);
                        } else {
                             // Buffer full, maybe beep? Ignore char for now.
                             MessageBeep(MB_ICONASTERISK); // Simple feedback
                        }
                    }
                    // Ignore other keys (arrows, function keys, etc.) for now
                } // End if KEY_EVENT down
            } // End for loop processing records
        } // End if WAIT_OBJECT_0
        else if (waitResult == WAIT_OBJECT_0 + 1) { // Shutdown event
            final_status = FAILURE;
            goto cleanup;
        }
         else { // Wait failed or returned unexpected value
             print_stderr("WaitForMultipleObjects failed or returned unexpected value (%lu, Error %lu).", waitResult, GetLastError());
             final_status = ERR_CALCULATION;
             goto cleanup;
        }
    } // End while loop

    // If loop finishes because buffer is full without Enter
    if (buffer_pos >= buffer_size - 1) {
         // Echo newline for consistency before error
         DWORD written;
         WriteConsoleA(hStdout, "\r\n", 2, &written, NULL);
         fprintf(stderr, "\nInput line too long (max %zu characters). Input discarded.\n\n", buffer_size - 2);
         fflush(stderr);
         // We don't have an easy way to clear the *rest* of the user's typed input here
         // with ReadConsoleInput, as it was never added to our buffer.
         // FlushConsoleInputBuffer(hStdin); // Could potentially discard unprocessed events
         final_status = ERR_INVALID_INPUT;
    }

cleanup:
    // --- Restore initial console mode ---
    if (!SetConsoleMode(hStdin, initial_mode)) {
        // Don't overwrite primary error, but report this one too
        print_stderr("Failed to restore initial console mode (Error %lu).", GetLastError());
        if (final_status == SUCCESS || final_status == FAILURE) {
             final_status = ERR_CALCULATION; // Promote to error if restoration failed
        }
    }
    return final_status;
}


#else // --- Implementation for POSIX ---

// ... (read_line_fgets - unchanged from previous correct version) ...
/**
 * @brief Reads a line robustly using fgets, handles errors, shutdown, and buffer overflow. (POSIX version)
 */
static StatusCode read_line_fgets(const char* prompt, char* buffer, size_t buffer_size) {
     if (prompt) {
        printf("%s", prompt);
        fflush(stdout); // Ensure prompt is visible
    }

    // Check flag *before* blocking call
    if (g_shutdown_requested) {
        return FAILURE;
    }

    clearerr(stdin);
    errno = 0; // Clear errno before calling fgets

    if (!fgets(buffer, (int)buffer_size, stdin)) {
        // Handle fgets failure
        int fgets_errno = errno; // Capture errno immediately

        // Check 1: Was shutdown requested *during* or immediately *after* fgets attempted?
        if (g_shutdown_requested) {
            return FAILURE; // Shutdown detected
        }

        // Check 2: Was fgets interrupted by a signal (and SA_RESTART is off)?
        if (fgets_errno == EINTR) {
            clearerr(stdin);
            // Return SUCCESS, caller loop MUST check g_shutdown_requested
            return SUCCESS;
        }

        // Check 3: Handle EOF explicitly (map to ERR_CALCULATION)
        if (feof(stdin)) {
            clearerr(stdin);
            return ERR_CALCULATION; // Map EOF to calculation error
        }

        // Check 4: Other read errors
        print_stderr("Failed reading input: %s (errno %d)", strerror(fgets_errno), fgets_errno);
        clearerr(stdin);
        return ERR_CALCULATION; // Return specific error code
    }

    // Check 5: Flag set immediately after successful read?
    if (g_shutdown_requested) {
        return FAILURE;
    }

    // --- Check for buffer overflow & remove newline ---
    char* newline_ptr = strchr(buffer, '\n');
    if (newline_ptr == NULL) {
        size_t len = strlen(buffer);
        if (len == buffer_size - 1 && buffer[len-1] != '\n') {
            fprintf(stderr, "\nInput line too long (max %zu characters). Input discarded.\n\n", buffer_size - 2);
            fflush(stderr);
            clear_stdin_buffer();
            return ERR_INVALID_INPUT;
        }
    } else {
        *newline_ptr = '\0';
    }

    return SUCCESS; // Successful read
}

#endif // _WIN32


// --- Input Gathering Function Implementations ---
// These MUST be updated to call the correct new Windows function

/**
 * @brief Gets a positive double value from the user.
 */
StatusCode get_positive_double(const char* prompt, double* value) {
    char buffer[MAX_INPUT_BUFFER_SIZE];
    char* endptr;
    if (!value) return ERR_INVALID_INPUT;

    while (true) {
        if (g_shutdown_requested) { return FAILURE; }

        // Platform-specific read call
        StatusCode read_status;
        #ifdef _WIN32
            read_status = read_line_windows_raw(prompt, buffer, sizeof(buffer)); // USE NEW FUNCTION
        #else
            read_status = read_line_fgets(prompt, buffer, sizeof(buffer));
            if (read_status == SUCCESS && g_shutdown_requested) { return FAILURE; }
        #endif

        if (g_shutdown_requested) { return FAILURE; }

        if (read_status == FAILURE) { return FAILURE; }
        if (read_status == ERR_INVALID_INPUT) { continue; }
        if (read_status == ERR_CALCULATION) { return ERR_CALCULATION; }

        if (read_status == SUCCESS) {
             // Parsing and Validation Logic (remains unchanged)
             // ... (same as before) ...
              char* start = buffer;
             while (isspace((unsigned char)*start)) start++;

             if (*start == '\0') {
                 fprintf(stderr, "\nNo input provided.\n\n"); fflush(stderr);
                 continue; // Re-prompt
             }

             errno = 0; // Reset for strtod
             *value = strtod(start, &endptr);
             int strtod_errno = errno;
             bool conversion_ok = true;

             // --- Validation Logic ---
             if (strtod_errno == ERANGE) {
                 fprintf(stderr, "\nInput value is out of range for a double.\n\n"); fflush(stderr);
                 conversion_ok = false;
             } else if (strtod_errno != 0 && *value == 0) {
                 fprintf(stderr, "\nInput conversion failed: %s (errno %d)\n\n", strerror(strtod_errno), strtod_errno); fflush(stderr);
                 conversion_ok = false;
             } else if (endptr == start) {
                 fprintf(stderr, "\nNo valid number found in input.\n\n"); fflush(stderr);
                 conversion_ok = false;
             } else {
                 char* check_ptr = endptr;
                 while (*check_ptr != '\0' && isspace((unsigned char)*check_ptr)) check_ptr++;
                 if (*check_ptr != '\0') {
                     fprintf(stderr, "\nTrailing characters found after number: '%s'\n\n", check_ptr); fflush(stderr);
                     conversion_ok = false;
                 }
                 // Application-specific validation
                 else if (*value <= 0.0) { fprintf(stderr, "\nValue must be positive.\n\n"); fflush(stderr); conversion_ok = false; }
                 else if (!isfinite(*value)) { fprintf(stderr, "\nValue must be finite (not NaN or infinity).\n\n"); fflush(stderr); conversion_ok = false; }
                 else if (*value > MAX_REASONABLE_RATE) { fprintf(stderr, "\nInput rate is unreasonably large (exceeds %.1f THz).\n\n", MAX_REASONABLE_RATE / 1.0e12); fflush(stderr); conversion_ok = false; }
                 else if (*value < MIN_REASONABLE_RATE) { fprintf(stderr, "\nInput rate is unreasonably small (less than %.1f Hz).\n\n", MIN_REASONABLE_RATE); fflush(stderr); conversion_ok = false; }

             } // End validation checks

             if (conversion_ok) {
                 return SUCCESS; // Valid input found
             }
             // If conversion failed, loop continues
             continue;
        }
    }
}

// ... (Similarly update get_specific_integer, get_positive_long_long, get_confirmation,
//      and the inner loops in get_user_inputs to call read_line_windows_raw on Windows) ...


/**
 * @brief Gets an integer that must be one of the allowed values.
 */
StatusCode get_specific_integer(const char* prompt, const int allowed_values[], int num_allowed, int* value) {
     char buffer[MAX_INPUT_BUFFER_SIZE];
     char* endptr;
     long long ll_val;
     if (!value || !allowed_values || num_allowed <= 0) return ERR_INVALID_INPUT;

     while(true) {
         if (g_shutdown_requested) { return FAILURE; }

         StatusCode read_status;
        #ifdef _WIN32
            read_status = read_line_windows_raw(prompt, buffer, sizeof(buffer)); // USE NEW FUNCTION
        #else
            read_status = read_line_fgets(prompt, buffer, sizeof(buffer));
            if (read_status == SUCCESS && g_shutdown_requested) { return FAILURE; }
        #endif

         if (g_shutdown_requested) { return FAILURE; }

         if (read_status == FAILURE) { return FAILURE; }
         if (read_status == ERR_INVALID_INPUT) { continue; }
         if (read_status == ERR_CALCULATION) { return ERR_CALCULATION; }

         if (read_status == SUCCESS) {
             // --- Parsing & Validation (Unchanged) ---
            char* start = buffer;
             while(isspace((unsigned char)*start)) start++;

             if (*start == '\0') {
                 fprintf(stderr, "\nNo input provided.\n\n"); fflush(stderr);
                 continue; // Re-prompt
             }

             errno = 0; // Reset for strtoll
             ll_val = strtoll(start, &endptr, 10);
             int strtoll_errno = errno;
             bool conversion_ok = true;

             // Validation Logic
             if (strtoll_errno == ERANGE) {
                 fprintf(stderr, "\nInput value is out of range for an integer.\n\n"); fflush(stderr);
                 conversion_ok = false;
             } else if (strtoll_errno != 0 && ll_val == 0) {
                 fprintf(stderr, "\nInput conversion failed: %s (errno %d)\n\n", strerror(strtoll_errno), strtoll_errno); fflush(stderr);
                 conversion_ok = false;
             } else if (endptr == start) {
                 fprintf(stderr, "\nNo valid integer found in input.\n\n"); fflush(stderr);
                 conversion_ok = false;
             } else {
                 char* check_ptr = endptr;
                 while (*check_ptr != '\0' && isspace((unsigned char)*check_ptr)) check_ptr++;
                 if (*check_ptr != '\0') {
                     fprintf(stderr, "\nTrailing characters found after integer: '%s'\n\n", check_ptr); fflush(stderr);
                     conversion_ok = false;
                 }
                 else if (ll_val < INT_MIN || ll_val > INT_MAX) {
                     fprintf(stderr, "\nInput value is outside the valid range for the required integer type.\n\n"); fflush(stderr);
                     conversion_ok = false;
                 } else {
                     *value = (int)ll_val;
                     bool found = false;
                     for (int i = 0; i < num_allowed; ++i) {
                         if (*value == allowed_values[i]) {
                             found = true;
                             break;
                         }
                     }
                     if (!found) {
                         fprintf(stderr, "\nError: Input must be one of the allowed values: ");
                         for (int i = 0; i < num_allowed; ++i) {
                             fprintf(stderr, "%d%s", allowed_values[i], (i < num_allowed - 1) ? ", " : ".\n\n");
                         }
                         fflush(stderr);
                         conversion_ok = false;
                     }
                 }
             } // End validation checks

             if(conversion_ok) {
                 return SUCCESS; // Found valid, allowed integer
             }
             continue;
         }
     }
}


/**
 * @brief Gets a positive long long value from the user.
 */
StatusCode get_positive_long_long(const char* prompt, long long* value) {
    char buffer[MAX_INPUT_BUFFER_SIZE];
    char* endptr;
    if (!value) return ERR_INVALID_INPUT;

    while(true) {
         if (g_shutdown_requested) return FAILURE;

        StatusCode read_status;
        #ifdef _WIN32
            read_status = read_line_windows_raw(prompt, buffer, sizeof(buffer)); // USE NEW FUNCTION
        #else
            read_status = read_line_fgets(prompt, buffer, sizeof(buffer));
            if (read_status == SUCCESS && g_shutdown_requested) { return FAILURE; }
        #endif

        if (g_shutdown_requested) return FAILURE;

        if (read_status == FAILURE) { return FAILURE; }
        if (read_status == ERR_INVALID_INPUT) { continue; }
        if (read_status == ERR_CALCULATION) { return ERR_CALCULATION; }

        if (read_status == SUCCESS) {
             // --- Parsing & Validation (Unchanged) ---
            char* start = buffer;
            while(isspace((unsigned char)*start)) start++;

            if (*start == '\0') {
                fprintf(stderr, "\nNo input provided.\n\n"); fflush(stderr);
                 continue; // Re-prompt
            }

            errno = 0; // Reset for strtoll
            *value = strtoll(start, &endptr, 10);
            int strtoll_errno = errno;
            bool conversion_ok = true;

            // Validation Logic
            if (strtoll_errno == ERANGE) {
                if (*value == LLONG_MAX) fprintf(stderr, "\nInput value is too large (overflow).\n\n");
                else if (*value == LLONG_MIN) fprintf(stderr, "\nInput value is too small (underflow).\n\n");
                else fprintf(stderr, "\nInput value is out of range for a long long integer.\n\n");
                fflush(stderr);
                conversion_ok = false;
            } else if (strtoll_errno != 0 && *value == 0) {
                fprintf(stderr, "\nInput conversion failed: %s (errno %d)\n\n", strerror(strtoll_errno), strtoll_errno); fflush(stderr);
                conversion_ok = false;
            } else if (endptr == start) {
                fprintf(stderr, "\nNo valid integer found in input.\n\n"); fflush(stderr);
                conversion_ok = false;
            } else {
                 char* check_ptr = endptr;
                 while (*check_ptr != '\0' && isspace((unsigned char)*check_ptr)) check_ptr++;
                 if (*check_ptr != '\0') {
                     fprintf(stderr, "\nTrailing characters found after integer: '%s'\n\n", check_ptr); fflush(stderr);
                     conversion_ok = false;
                 }
                 else if (*value <= 0) {
                     fprintf(stderr, "\nValue must be positive.\n\n"); fflush(stderr);
                     conversion_ok = false;
                 }
            } // End validation checks

            if(conversion_ok) {
                return SUCCESS; // Valid positive long long found
            }
             continue;
        }
    }
}


/**
 * @brief Gets a yes/no confirmation from the user. Defaults to 'no' on Enter.
 */
bool get_confirmation(const char* prompt) {
    char buffer[MAX_INPUT_BUFFER_SIZE];
    while(true) {
        if (g_shutdown_requested) return false;

        StatusCode read_status;
        #ifdef _WIN32
            read_status = read_line_windows_raw(prompt, buffer, sizeof(buffer)); // USE NEW FUNCTION
        #else
            read_status = read_line_fgets(prompt, buffer, sizeof(buffer));
            if (read_status == SUCCESS && g_shutdown_requested) { return false; }
        #endif

        if (g_shutdown_requested) return false;

        if (read_status == FAILURE) { return false; }
        if (read_status == ERR_INVALID_INPUT) { continue; }
        if (read_status == ERR_CALCULATION) { continue; }

        if (read_status == SUCCESS) {
            // --- Process if read_status == SUCCESS (Unchanged) ---
             char* start = buffer;
            while (isspace((unsigned char)*start)) start++;

            char* end = start + strlen(start);
            while (end > start && isspace((unsigned char)*(end - 1))) end--;
            *end = '\0';

            if (*start == '\0') {
                return false; // Empty input defaults to 'no'
            } else if (strlen(start) == 1) {
                 char lower_char = tolower((unsigned char)*start);
                 if (lower_char == 'y') { return true; }
                 else if (lower_char == 'n') { return false; }
            }

            fprintf(stderr, "\nInvalid input '%s'. Please enter 'y', 'n', or press Enter for the default ('n').\n\n", buffer);
            fflush(stderr);
            continue;
        }
    }
}


/**
 * @brief Gathers all necessary user inputs for the program configuration.
 */
StatusCode get_user_inputs(UserInputs* inputs) {
    if (!inputs) return ERR_INVALID_INPUT;
    StatusCode status = SUCCESS;
    StatusCode ret_status;

    memset(inputs, 0, sizeof(UserInputs));
    inputs->allow_m1_greater_than_l1 = false;
    inputs->intermediate_rate_targeted = false;
    inputs->target_intermediate_rate = 0.0;
    inputs->intermediate_priority = PRIORITIZE_BALANCED_DUAL;
    inputs->max_lm_per_stage = 0;

    status = get_positive_double("Enter the input sample rate (Hz): ", &inputs->input_rate);
    if (g_shutdown_requested || status != SUCCESS) {
        ret_status = (status == SUCCESS) ? FAILURE : status;
        return ret_status;
    }

    status = get_positive_double("Enter the output sample rate (Hz): ", &inputs->output_rate);
    if (g_shutdown_requested || status != SUCCESS) {
        ret_status = (status == SUCCESS) ? FAILURE : status;
        return ret_status;
    }

    int allowed_modes[] = {ONE_RESAMPLER, TWO_RESAMPLERS};
    int mode_choice;
    status = get_specific_integer("Use one or two rational resamplers? (1 or 2): ", allowed_modes, 2, &mode_choice);
    if (g_shutdown_requested || status != SUCCESS) {
        ret_status = (status == SUCCESS) ? FAILURE : status;
        return ret_status;
    }
    inputs->num_resamplers = (ResamplerMode)mode_choice;

    if (inputs->num_resamplers == TWO_RESAMPLERS) {
        printf("\n");
        printf("Two-Resampler Search Mode:\n\n");
        printf(" 1: Default Minimal Sum (Find match/approximation with smallest L1+M1+L2+M2 sum)\n");
        printf(" 2: Custom Limit Search (Search using custom max L/M values you provide)\n");
        printf("\n");

        while (true) {
            if (g_shutdown_requested) return FAILURE;
            printf("Select search mode (1 or 2) [Enter=1]: ");
            fflush(stdout);

            char buffer[MAX_INPUT_BUFFER_SIZE];
            StatusCode read_status;
            #ifdef _WIN32
                read_status = read_line_windows_raw(NULL, buffer, sizeof(buffer)); // USE NEW FUNCTION
            #else
                read_status = read_line_fgets(NULL, buffer, sizeof(buffer));
                if (read_status == SUCCESS && g_shutdown_requested) { return FAILURE; }
            #endif

            if (g_shutdown_requested) return FAILURE;

            if (read_status == FAILURE) return FAILURE;
            if (read_status == ERR_INVALID_INPUT) { continue; }
            if (read_status == ERR_CALCULATION) { return ERR_CALCULATION; }

            // --- Parsing (Unchanged) ---
            char* start = buffer;
            while (isspace((unsigned char)*start)) start++;

            if (*start == '\0') {
                inputs->search_mode = SEARCH_MINIMAL_SUM;
                printf(" -> Minimal Sum mode selected (default).\n");
                break;
            }

            char* endptr;
            errno = 0;
            long long ll_val = strtoll(start, &endptr, 10);
            int strtoll_errno = errno;
            bool conversion_ok = true;
            int search_choice = 0;

            // Validation (Unchanged)
            if (strtoll_errno == ERANGE) { fprintf(stderr, "\nInput out of range.\n\n"); fflush(stderr); conversion_ok = false; }
            else if (strtoll_errno != 0 && ll_val == 0) { fprintf(stderr, "\nConversion failed: %s\n\n", strerror(strtoll_errno)); fflush(stderr); conversion_ok = false; }
            else if (endptr == start) { fprintf(stderr, "\nNo integer found.\n\n"); fflush(stderr); conversion_ok = false; }
            else {
                char* check_ptr = endptr;
                while (isspace((unsigned char)*check_ptr)) check_ptr++;
                if (*check_ptr != '\0') { fprintf(stderr, "\nTrailing characters: '%s'\n\n", check_ptr); fflush(stderr); conversion_ok = false; }
                else if (ll_val < INT_MIN || ll_val > INT_MAX) { fprintf(stderr, "\nValue out of range for int.\n\n"); fflush(stderr); conversion_ok = false; }
                else {
                    search_choice = (int)ll_val;
                    if (search_choice != 1 && search_choice != 2) {
                        fprintf(stderr, "\nInput must be 1 or 2.\n\n"); fflush(stderr);
                        conversion_ok = false;
                    }
                }
            }

            if (conversion_ok) {
                inputs->search_mode = (search_choice == 2) ? SEARCH_DEFAULT : SEARCH_MINIMAL_SUM;
                printf(" -> %s mode selected.\n", (inputs->search_mode == SEARCH_MINIMAL_SUM) ? "Minimal Sum" : "Custom Limit Search");
                break;
            } else {
                continue;
            }
        }


        // Get Max L/M Limit (Unchanged)
        if (inputs->search_mode == SEARCH_MINIMAL_SUM) {
             inputs->max_lm_per_stage = HARD_MAX_LM_LIMIT;
             printf("\nNote: Minimal Sum mode selected. Using internal max L/M limit: %lld.\n", inputs->max_lm_per_stage);
             printf("      and explores solutions for the lowest L+M sum.\n");
        } else {
            long long current_attempt_max_lm = 0;
            bool confirmed_large = false;
            while (true) {
                 printf("\n");
                 status = get_positive_long_long("Enter maximum L and M for EACH stage (e.g., 200): ", &current_attempt_max_lm);
                 if (g_shutdown_requested || status != SUCCESS) {
                     ret_status = (status == SUCCESS) ? FAILURE : status;
                     return ret_status;
                 }

                 if (current_attempt_max_lm > HARD_MAX_LM_LIMIT) {
                      fprintf(stderr, "\nMaximum L/M value (%lld) exceeds the hard limit of %lld.\nPlease use a value <= %lld.\n\n",
                                   current_attempt_max_lm, HARD_MAX_LM_LIMIT, HARD_MAX_LM_LIMIT );
                      fflush(stderr);
                      continue;
                 }
                 if (!confirmed_large && current_attempt_max_lm > LARGE_LM_THRESHOLD) {
                     printf("\nWarning: Maximum L/M per stage = %lld is large.\n", current_attempt_max_lm);
                     printf("Resamplers using large L/M values are typically more complex and computationally expensive.\n\n");
                     bool confirmed = get_confirmation("Continue with this value? (y/n) [Enter=n]: ");
                     if (g_shutdown_requested) { return FAILURE; }
                     if (!confirmed) {
                          printf("Operation cancelled by user. Please enter a different value.\n");
                          continue;
                     }
                     confirmed_large = true;
                 }
                 inputs->max_lm_per_stage = current_attempt_max_lm;
                 break;
            }
        }


        // Get M1 > L1 Constraint (Unchanged)
        printf("\n"); printf("Constraint: By default, the first stage ensures L1 >= M1.\n");
        printf("            Allow the first stage to downsample (M1 > L1)?\n");
        printf("            (Useful for intermediate processing at lower rates, but may seem\n");
        printf("            less direct for pure sample rate conversion.)\n"); printf("\n");
        inputs->allow_m1_greater_than_l1 = get_confirmation("Allow M1 > L1? (y/n) [Enter=n]: ");
        if (g_shutdown_requested) { return FAILURE; }


        // Get Optional Intermediate Rate Target (Unchanged)
        printf("\n"); printf("Optional: Specify a target intermediate sample rate.\n");
        printf("          If skipped (by pressing Enter), the search will not target\n");
        printf("          a specific intermediate rate.\n"); printf("\n");

        while(true) {
            if (g_shutdown_requested) return FAILURE;
            printf("Enter target intermediate rate (Hz) (rate after L1/M1) [Press Enter to skip]: ");
            fflush(stdout);

            char buffer_ir[MAX_INPUT_BUFFER_SIZE];
            StatusCode read_status_ir;
            #ifdef _WIN32
                read_status_ir = read_line_windows_raw(NULL, buffer_ir, sizeof(buffer_ir)); // USE NEW FUNCTION
            #else
                read_status_ir = read_line_fgets(NULL, buffer_ir, sizeof(buffer_ir));
                if (read_status_ir == SUCCESS && g_shutdown_requested) { return FAILURE; }
            #endif

            if (g_shutdown_requested) return FAILURE;

            if (read_status_ir == FAILURE) return FAILURE;
            if (read_status_ir == ERR_INVALID_INPUT) { continue; }
            if (read_status_ir == ERR_CALCULATION) { return ERR_CALCULATION; }

             // --- Parse if SUCCESS (Unchanged) ---
             char* start_ir = buffer_ir;
             while (isspace((unsigned char)*start_ir)) start_ir++;

            if (*start_ir == '\0') {
                inputs->intermediate_rate_targeted = false;
                inputs->target_intermediate_rate = 0.0;
                inputs->intermediate_priority = PRIORITIZE_BALANCED_DUAL;
                printf(" -> Skipped intermediate rate target.\n");
                break;
            }

            char* endptr_ir;
            errno = 0;
            double temp_val = strtod(start_ir, &endptr_ir);
            int strtod_errno_ir = errno;
            bool conversion_ok_ir = true;

            // Validation (Unchanged)
            if (strtod_errno_ir == ERANGE) { fprintf(stderr, "\nInput out of range for double.\n\n"); fflush(stderr); conversion_ok_ir = false; }
            else if (strtod_errno_ir != 0 && temp_val == 0) { fprintf(stderr, "\nInput conversion failed: %s\n\n", strerror(strtod_errno_ir)); fflush(stderr); conversion_ok_ir = false; }
            else if (endptr_ir == start_ir) { fprintf(stderr, "\nNo number found in input.\n\n"); fflush(stderr); conversion_ok_ir = false; }
            else {
                char* check_ptr_ir = endptr_ir;
                while (*check_ptr_ir != '\0' && isspace((unsigned char)*check_ptr_ir)) check_ptr_ir++;
                if (*check_ptr_ir != '\0') { fprintf(stderr, "\nTrailing characters found after number: '%s'\n\n", check_ptr_ir); fflush(stderr); conversion_ok_ir = false; }
                else if (temp_val <= 0.0) { fprintf(stderr, "\nTarget rate must be positive.\n\n"); fflush(stderr); conversion_ok_ir = false; }
                else if (!isfinite(temp_val)) { fprintf(stderr, "\nTarget rate must be finite.\n\n"); fflush(stderr); conversion_ok_ir = false; }
                else if (temp_val < MIN_REASONABLE_RATE) { fprintf(stderr, "\nTarget rate is too small (less than %.1f Hz).\n\n", MIN_REASONABLE_RATE); fflush(stderr); conversion_ok_ir = false; }
                else if (temp_val > MAX_REASONABLE_RATE) { fprintf(stderr, "\nTarget rate is too large (exceeds %.1f THz).\n\n", MAX_REASONABLE_RATE / 1.0e12); fflush(stderr); conversion_ok_ir = false; }
            }

             if (conversion_ok_ir) {
                 inputs->intermediate_rate_targeted = true;
                 inputs->target_intermediate_rate = temp_val;
                 printf(" -> Targeting intermediate rate: %.2f Hz\n", temp_val);
                 break;
             } else {
                 continue;
             }
        }


        // Intermediate Rate Warning (Unchanged)
        if (inputs->intermediate_rate_targeted && inputs->target_intermediate_rate > inputs->input_rate && inputs->target_intermediate_rate > inputs->output_rate) {
             printf("\nWarning: Target intermediate rate (%.2f Hz) is higher than both input (%.2f Hz)\n", inputs->target_intermediate_rate, inputs->input_rate);
             printf("         and output (%.2f Hz). This implies significant oversampling then\n", inputs->output_rate);
             printf("         downsampling. While valid, this is less common.\n\n");
             fflush(stdout);
             bool confirm_intermediate = get_confirmation("Proceed with this intermediate target? (y/n) [Enter=n]: ");
             if (g_shutdown_requested) { return FAILURE; }
             if (!confirm_intermediate) {
                 printf("Operation cancelled by user due to intermediate rate configuration.\n");
                 return FAILURE;
             }
             printf(" -> Proceeding with specified intermediate rate.\n");
        }

        // Intermediate Rate Prioritization Prompt (Unchanged)
        if (inputs->intermediate_rate_targeted) {
             printf("\nIntermediate Rate Prioritization:\n\n");
             printf("How should the search prioritize finding this rate versus the final output rate?\n\n");
             printf(" 1: Strictly Enforce Intermediate Rate:\n");
             printf("    Only solutions matching the intermediate target (within %.1e relative error)\n    will be considered. The best of these is chosen based on the search mode\n    (Minimal Sum or best overall accuracy).\n", INTERMEDIATE_MATCH_TOLERANCE);
             printf(" 2: Prioritize Final Output Rate:\n");
             printf("    Solutions matching the final target (within %.1e relative error) are preferred.\n    Among those, the one closest to the intermediate target is chosen.\n    If no final match exists, the overall best approximation (by search mode) is chosen.\n", MATCH_TOLERANCE);
             printf(" 3: Balanced - Best Dual Match (if possible):\n");
             printf("    Solutions matching *both* targets are preferred. Among those, the best by\n    search mode is chosen. If no dual match exists, falls back to\n    PRIORITIZE_FINAL_RATE logic.\n");
             printf(" 4: Ignore Intermediate Rate for Selection:\n");
             printf("    The intermediate target is ignored during selection. The best solution\n    is chosen based *only* on the final rate accuracy and the primary search mode.\n    The achieved intermediate rate is still reported.\n");
             printf("\n");

             int allowed_priorities[] = { PRIORITIZE_STRICT_INTERMEDIATE, PRIORITIZE_FINAL_RATE, PRIORITIZE_BALANCED_DUAL, PRIORITIZE_BEST_OVERALL };
             int priority_choice;
             status = get_specific_integer("Select prioritization mode (1-4): ", allowed_priorities, 4, &priority_choice);
             if (g_shutdown_requested || status != SUCCESS) {
                 ret_status = (status == SUCCESS) ? FAILURE : status;
                 return ret_status;
             }
             inputs->intermediate_priority = (IntermediatePriority)priority_choice;
        }
    } else { // ONE_RESAMPLER case (Unchanged)
        inputs->search_mode = SEARCH_MINIMAL_SUM;
        inputs->allow_m1_greater_than_l1 = false;
        inputs->intermediate_rate_targeted = false;
        inputs->target_intermediate_rate = 0.0;
        inputs->intermediate_priority = PRIORITIZE_BALANCED_DUAL;
        inputs->max_lm_per_stage = HARD_MAX_LM_LIMIT;
    }

    return SUCCESS;
} // End get_user_inputs
