// src/main.c

#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include "utils.h"
#include "signal_handler.h" // Needed for g_shutdown_requested and setup/cleanup
#include "input.h"
#include "single_resampler.h"
#include "two_resamplers.h"

// Include necessary header for isatty() and Windows handle cleanup
#ifdef _WIN32
    #include <io.h>     // For _isatty on Windows (MSVCRT)
    #include <windows.h> // For CloseHandle <-- Needed for cleanup
    #define ISATTY _isatty // Define a common name
    #define FILENO _fileno // Define a common name
#else
    #include <unistd.h> // For isatty on POSIX
    #define ISATTY isatty
    #define FILENO fileno
#endif


int main(void) {
    UserInputs inputs;
    StatusCode last_op_status = SUCCESS; // Track status of last major operation

    // --- Check if running interactively (Unchanged) ---
    int stdin_fd = FILENO(stdin);
    if (stdin_fd < 0) {
         fprintf(stderr, "Error: Could not get file descriptor for stdin.\n");
         return EXIT_FAILURE;
    }
    if (!ISATTY(stdin_fd)) {
        fprintf(stderr, "Error: This program requires interactive input from a terminal.\n");
        fprintf(stderr, "       Input redirection or piping is not supported.\n");
        return EXIT_FAILURE;
    }
    // --- End interactive check ---


    // --- Set Locale (Unchanged) ---
    if (setlocale(LC_NUMERIC, "C") == NULL) {
         fprintf(stderr, "\nWarning: Could not set C locale for numeric input/output. Parsing might fail.\n\n");
    }

    // --- Setup signal handlers (Unchanged - calls the updated setup) ---
    // This now also creates the shutdown event on Windows
    if (!setup_signal_handlers()) {
        // Message already printed by setup_signal_handlers on failure
        fprintf(stderr, "Error: Failed to initialize signal handling. Exiting.\n");
        #ifdef _WIN32
        // Ensure event handle is cleaned up if it was partially created before failure
        if (g_hShutdownEvent != NULL) {
            CloseHandle(g_hShutdownEvent);
            g_hShutdownEvent = NULL;
        }
        #endif
        return EXIT_FAILURE;
    }

    // --- Print Program Header (Unchanged) ---
    printf("\n=== Rational Resampler L/M Calculator ===\n");
    printf("Calculates Interpolation (L) and Decimation (M) values.\n");
    printf("Uses Continued Fraction Expansion (CFE) and Factorization.\n");
    printf("=====================================================\n\n");

    // --- Main Application Loop (Unchanged - relies on updated input functions) ---
    while (!g_shutdown_requested) {

        // Get user input parameters
        // This function now uses the appropriate interruptible input method
        last_op_status = get_user_inputs(&inputs);

        // Check flag immediately after potentially blocking input call returns
        if (g_shutdown_requested) {
            if (last_op_status == SUCCESS) last_op_status = FAILURE;
            break;
        }

        // Handle input function return status (Unchanged)
        if (last_op_status != SUCCESS) {
            if (!g_shutdown_requested) {
                 // Use print_stderr for consistent error formatting if it wasn't shutdown
                 print_stderr("Failed to gather required inputs (%s).", status_code_to_string(last_op_status));
            } else {
                 // Shutdown requested, message handled later.
            }
            break; // Exit loop on any input failure
        }


        // --- Perform Calculations based on input (Unchanged) ---
        if (inputs.num_resamplers == ONE_RESAMPLER) {
            last_op_status = handle_one_resampler(&inputs);
        } else { // TWO_RESAMPLERS
            last_op_status = prompt_and_run_two_resamplers(&inputs);
        }

        // Check flag immediately after calculation finishes (Unchanged)
        if (g_shutdown_requested) {
            if (last_op_status == SUCCESS) last_op_status = FAILURE; // Mark as failure if interrupted
            break;
        }

        // Handle calculation results (Unchanged)
        if (last_op_status < FAILURE && last_op_status != SUCCESS) {
             print_stderr("Calculation phase failed with error: %s", status_code_to_string(last_op_status));
             break; // Exit loop on critical errors
        }

        // --- Ask to Continue (Unchanged - uses updated input function) ---
        if (last_op_status == SUCCESS || last_op_status == FAILURE) {
            printf("\n");
            fflush(stdout);
            // get_confirmation handles its own interrupt checks
            bool continue_loop = get_confirmation("Perform another sample rate conversion? (y/n) [Enter=n]: ");

            // Check flag immediately after get_confirmation returns
            if (g_shutdown_requested) {
                 if (last_op_status == SUCCESS) last_op_status = FAILURE;
                 break; // Exit main loop
            }

            if (!continue_loop) {
                if (last_op_status != FAILURE) last_op_status = SUCCESS;
                break; // Exit main loop
            }
            printf("\n"); // Spacing for next iteration
        }

    } // End main application loop


    // --- Cleanup and Exit ---
    if (g_shutdown_requested) {
         print_stderr("\nShutdown requested by user.");
         if (last_op_status == SUCCESS) {
             last_op_status = FAILURE;
         }
    }

    // Determine final exit code (Unchanged)
    int exit_code = (last_op_status == SUCCESS) ? EXIT_SUCCESS : EXIT_FAILURE;

    // Print final message (Unchanged)
    if (exit_code == EXIT_SUCCESS) {
        printf("\nExiting program normally.\n");
    } else {
        if (last_op_status != FAILURE || !g_shutdown_requested) {
             printf("\nExiting program due to error or user request.\n");
        } else {
             printf("\n"); // Just a newline if it was a shutdown request
        }
    }
    fflush(stdout);


    // --- Platform-Specific Cleanup ---
    #ifdef _WIN32
    // Close the shutdown event handle created in setup_signal_handlers
    if (g_hShutdownEvent != NULL) {
        CloseHandle(g_hShutdownEvent);
        g_hShutdownEvent = NULL; // Good practice to NULL the handle after closing
    }
    #endif

    return exit_code;
}
