// src/signal_handler.c

#include "signal_handler.h" // Include own header first
#include "utils.h"          // For print_stderr (needed for warnings)
#include <signal.h>  // For sig_atomic_t, sigaction, SIGINT etc.
#include <stdbool.h> // For bool return type
#include <string.h> // For memset

// Platform-specific includes for implementation
#ifdef _WIN32
#include <windows.h>
// NOTE: No other includes needed specifically for this change
#else
#include <unistd.h> // For sigaction etc.
#endif

// --- Global Variable Definition ---
// Define the actual global variable here.
volatile sig_atomic_t g_shutdown_requested = 0;

#ifdef _WIN32
// Define the Windows-specific shutdown event handle
HANDLE g_hShutdownEvent = NULL;
#endif


// --- Signal Handler Implementations ---
// These handlers should be minimal and only use async-signal-safe operations.
// Setting a volatile sig_atomic_t variable is safe.
// Signaling an event (SetEvent) is generally considered safe from a console handler.

#ifdef _WIN32
// Windows console event handler
static BOOL WINAPI ConsoleHandler(DWORD CEvent) {
    switch(CEvent) {
        case CTRL_C_EVENT:
        case CTRL_BREAK_EVENT:
            g_shutdown_requested = 1;
            // Signal the event to potentially interrupt WaitForMultipleObjects
            if (g_hShutdownEvent != NULL) {
                SetEvent(g_hShutdownEvent);
            }
            return TRUE; // Indicate handled
        case CTRL_CLOSE_EVENT:
        case CTRL_LOGOFF_EVENT:
        case CTRL_SHUTDOWN_EVENT:
            g_shutdown_requested = 1;
             // Signal the event to potentially interrupt WaitForMultipleObjects
            if (g_hShutdownEvent != NULL) {
                SetEvent(g_hShutdownEvent);
            }
            // Returning TRUE might give a tiny window for cleanup, but OS may terminate quickly.
            return TRUE;
    }
    return FALSE; // Not handled
}
#else
// POSIX SIGINT handler (compatible with sigaction)
static void sigint_handler(int sig) {
    (void)sig; // Mark parameter as unused if not needed
    g_shutdown_requested = 1;
    // No other operations allowed here!
    // On POSIX, the EINTR mechanism handles interruption.
}
#endif

// --- Public Function Implementation ---

/**
 * @brief Sets up the signal handlers.
 */
bool setup_signal_handlers(void) {
#ifdef _WIN32
    // Create the manual-reset event used for interrupting waits
    // Security attributes = NULL, Manual reset = TRUE, Initial state = FALSE (nonsignaled), Name = NULL
    g_hShutdownEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (g_hShutdownEvent == NULL) {
        print_stderr("Fatal Error: Could not create shutdown event object (Error %lu).", GetLastError());
        return false; // Critical failure
    }

    if (!SetConsoleCtrlHandler((PHANDLER_ROUTINE)ConsoleHandler, TRUE)) {
        print_stderr("Fatal Error: Could not set Windows console control handler (Error %lu).", GetLastError());
        // Cleanup the event if handler setup fails
        CloseHandle(g_hShutdownEvent);
        g_hShutdownEvent = NULL;
        return false; // Treat this as critical because the event mechanism depends on it.
    }
    return true; // Success
#else // POSIX
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa)); // Initialize struct
    sa.sa_handler = sigint_handler; // Assign our handler function
    sigemptyset(&sa.sa_mask);       // Do not block any signals during handler execution
    // SA_RESTART is *not* set, so blocking syscalls like read() are interrupted with EINTR.
    // Note: Standard I/O functions like fgets might have different behavior regarding EINTR.
    sa.sa_flags = 0;

    // Register handler for SIGINT (Ctrl+C)
    if (sigaction(SIGINT, &sa, NULL) == -1) {
        print_stderr("Warning: Could not set SIGINT handler using sigaction.");
        // Consider this non-fatal for now, but return false might be safer
        return false;
    }
    // Optional: Register handler for SIGTERM
    // if (sigaction(SIGTERM, &sa, NULL) == -1) {
    //     print_stderr("Warning: Could not set SIGTERM handler using sigaction.");
    //     // return false; // Decide if this is critical
    // }

    return true; // Success
#endif
}
