// src/single_resampler.c

#include "single_resampler.h" // Include own header first
#include "signal_handler.h"   // For g_shutdown_requested
#include "resampling_common.h"  // For constants like HARD_MAX_LM_LIMIT, MATCH_TOLERANCE etc.
#include "utils.h"            // For print_stderr, status codes, status_code_to_string, max_long_long
#include "cfe_utils.h"        // For CFE calculation
#include <stdio.h>
#include <stdlib.h> // For free()
#include <math.h>   // For fabs, NAN, isfinite, INFINITY, copysign
#include <limits.h> // For LLONG_MAX (used implicitly by CFE checks)


// --- Function Implementations ---

/**
 * @brief Calculates the best single-stage L/M using CFE.
 * @details See header for full description.
 */
StatusCode calculate_single_resampler(double input_rate, double output_rate, long long* l_out, long long* m_out, double* achieved_output_rate) {
    if (g_shutdown_requested) return FAILURE;

    // Validate output pointers
    if (!l_out || !m_out || !achieved_output_rate) {
        // Use fprintf for internal errors as print_stderr adds extra newlines
        fprintf(stderr, "Internal Error (calc_single): Output pointers are NULL.\n");
        return ERR_INVALID_INPUT;
    }
    // Initialize outputs safely
    *l_out = 0;
    *m_out = 0;
    *achieved_output_rate = NAN;

    // Validate rate values (basic checks, main validation in input.c)
    if (!isfinite(input_rate) || !isfinite(output_rate) || input_rate <= 0 || output_rate <= 0) {
        print_stderr("Invalid rates for single resampler (must be positive finite).");
        return ERR_INVALID_INPUT;
    }

    // Calculate target ratio = Output / Input
    double target_ratio = output_rate / input_rate;
    if (!isfinite(target_ratio) || target_ratio <= 0) {
        print_stderr("Target ratio is invalid (non-positive or non-finite).");
        return ERR_CALCULATION;
    }

    // --- Use CFE to find best rational approximation L/M ---
    Convergent* convergents = NULL;
    int conv_count = 0;
    StatusCode cfe_status;
    // Max denominator M is capped by the hard limit for single stage
    long long max_denominator = HARD_MAX_LM_LIMIT;

    cfe_status = calculate_cfe_convergents(target_ratio, max_denominator, &convergents, &conv_count);

    // Handle CFE results
    if (cfe_status != SUCCESS) {
        if (cfe_status < FAILURE) { // Critical errors (Memory, Overflow etc.)
             print_stderr("Critical error during CFE calculation: %s.", status_code_to_string(cfe_status));
        } else { // Non-critical calculation issues
             if (conv_count == 0) {
                  print_stderr("CFE generated no convergents within M <= %lld for ratio %.6g.", max_denominator, target_ratio);
             } else {
                  // Could be overflow but partial results returned (cfe_status was SUCCESS)
                  // Or other ERR_CALCULATION from CFE internals.
                  // Let's assume CFE handled the specific message if needed.
                  // If status != SUCCESS and count > 0, it means overflow with partial results.
                  // CFE maps this to SUCCESS, so this branch won't be hit for that case.
                  print_stderr("CFE calculation failed or incomplete: %s.", status_code_to_string(cfe_status));
             }
             // Map calculation issues to ERR_CALCULATION for return
             cfe_status = ERR_CALCULATION;
        }
        // calculate_cfe_convergents handles its own memory cleanup on failure.
        return cfe_status; // Propagate CFE error status
    }

    // Check if CFE returned SUCCESS but found no convergents
    if (conv_count == 0) {
        // This case should have been mapped to ERR_CALCULATION inside CFE, but double-check.
        print_stderr("CFE generated no convergents within M <= %lld.", max_denominator);
        // No need to free convergents here (should be NULL or handled by CFE)
        return ERR_CALCULATION;
    }

    // --- Select the Best (Last) Convergent ---
    // CFE generates convergents in increasing order of complexity/accuracy.
    // The last one within the denominator limit is the best approximation.
    Convergent best_convergent = convergents[conv_count - 1];
    *l_out = best_convergent.p;
    *m_out = best_convergent.q;

    // Free the convergents array now we have the needed values
    free(convergents);
    convergents = NULL;

    // Final check on L and M values derived from CFE
    if (*m_out <= 0 || *l_out <= 0) {
        fprintf(stderr, "Internal Error: CFE resulted in non-positive L or M (%lld / %lld).\n", *l_out, *m_out);
        return ERR_CALCULATION;
    }

    // --- Calculate Achieved Rate ---
    // Use double precision for calculation to minimize intermediate overflow issues,
    // although L/M themselves are within long long limits.
    *achieved_output_rate = input_rate * ((double)*l_out / (double)*m_out);
    if (!isfinite(*achieved_output_rate)) {
        print_stderr("Achieved rate calculation resulted in non-finite value (L=%lld, M=%lld).", *l_out, *m_out);
        // Could be due to extreme L/M ratio or input_rate, even if L/M fit long long.
        return ERR_CALCULATION;
    }

    return SUCCESS;
}


/**
 * @brief Prints the results of the single-stage resampler calculation.
 * @details Format is preserved from original version. Uses max_long_long helper.
 */
void print_single_resampler_result(double input_rate, double output_rate, long long l, long long m, double achieved_output_rate) {
     if (g_shutdown_requested) return;

     // Basic validation of inputs for printing
     if (m <= 0 || l <= 0) { fprintf(stderr, "Internal Error (print_single): Cannot print result with L/M <= 0.\n"); return; }
     if (!isfinite(input_rate) || !isfinite(output_rate) || !isfinite(achieved_output_rate)) { fprintf(stderr, "Internal Error (print_single): Cannot print result with non-finite rates.\n"); return; }

     // Calculate differences and PPM for display
     double rate_diff_abs = fabs(achieved_output_rate - output_rate);
     // Use absolute diff if target is near zero to avoid large relative error/division by zero
     double rel_diff = (fabs(output_rate) > CFE_DBL_EPSILON) ? (rate_diff_abs / fabs(output_rate)) : rate_diff_abs;
     double ppm_error = NAN;
     if (fabs(output_rate) > CFE_DBL_EPSILON) {
         ppm_error = (achieved_output_rate - output_rate) / output_rate * 1000000.0;
     } else if (rate_diff_abs > CFE_DBL_EPSILON) {
          ppm_error = copysign(INFINITY, achieved_output_rate - output_rate); // Indicate direction if target is zero
     } else {
          ppm_error = 0.0;
     }


     // Determine status flags for notes
     bool is_match = (rel_diff < MATCH_TOLERANCE);
     bool significant_approx_error = !is_match && (rel_diff >= ARBITRARY_SUGGEST_ERROR_THRESHOLD);
     bool large_lm_factor = (l > SUGGEST_ARBITRARY_LM_THRESHOLD || m > SUGGEST_ARBITRARY_LM_THRESHOLD);

     // --- Print Results Title / Status ---
     // (Prompts/Structure unchanged)
     printf("\n");
     if (is_match && large_lm_factor) {
         printf("Results: Match Found with Large L/M\n\n");
     } else if (is_match) {
         printf("Results: Match Found!\n\n");
     } else if (significant_approx_error) {
         printf("Results: Caution! Significant Error\n\n");
     } else {
         printf("Results: Approximation Found\n\n");
     }

     // --- Print L/M Result ---
     // (Prompts unchanged)
     printf("  L = %lld (Interpolation)\n", l);
     printf("  M = %lld (Decimation)\n", m);

     // --- Calculate and Print Complexity and Achieved Rates ---
     // Use the consolidated helper function
     long long stage_max = max_long_long(l, m);
     double achieved_ratio = (double)l / (double)m; // Already know m > 0

     // (Prompts unchanged)
     printf("\nComplexity = %lld\n", stage_max); // Use max_long_long result
     printf("Achieved Ratio (L/M) = %lld / %lld = %.10f\n", l, m, achieved_ratio);
     printf("Achieved Output Rate = %.10f Hz\n", achieved_output_rate);
     printf("Rate Difference: %+.10f Hz, Relative: %.2e\n", achieved_output_rate - output_rate, rel_diff); // Show signed difference
     printf("PPM Error: ");
     if (isfinite(ppm_error)) {
         printf("%+.3f PPM\n", ppm_error);
     } else if (isnan(ppm_error)) {
          printf("N/A\n");
     } else { // +/- INFINITY
          printf("%sInfinity\n", (ppm_error > 0) ? "+" : "-");
     }


     // --- Notes Section ---
     // (Prompts and logic unchanged)
     if (significant_approx_error || large_lm_factor) {
         printf("\n"); // Blank line before notes

         if (significant_approx_error) {
             printf("Caution!: The best approximation found differs significantly\n");
             printf("          from the target ratio (Final Relative Error: %.2e).\n", rel_diff);
             printf("          An arbitrary (non-rational) resampling method might be required\n");
             printf("          for higher accuracy if this error is unacceptable.");
         }
         if (large_lm_factor) {
             if (significant_approx_error) printf("\n"); // Add spacing if both notes present
             printf("Note: One or more calculated factors (L=%lld, M=%lld) are large.\n", l, m);
             printf("      Implementing this rational ratio with standard polyphase filters may be\n");
             printf("      computationally expensive or require long filters.\n");
             printf("      Consider if an arbitrary resampling method might be a suitable\n");
             printf("      alternative if filter complexity is a primary concern.\n");
             printf("      Consider using the two-stage mode (option 2) to potentially find\n");
             printf("      factor pairs (L1/M1, L2/M2) with lower complexity.");
         }
         printf("\n"); // Blank line after notes
     }
     // --- End Notes Section ---

     fflush(stdout); // Ensure results are displayed
}


/**
 * @brief Handles the entire process for the single resampler mode.
 * @details Prints context, calls calculation, prints result/failure message.
 */
StatusCode handle_one_resampler(const UserInputs* inputs) {
    if (g_shutdown_requested) return FAILURE;
    if (!inputs) { fprintf(stderr, "Internal Error (handle_one): NULL inputs pointer.\n"); return ERR_INVALID_INPUT; }

    // --- Print Context ---
    // (Prompts unchanged)
    printf("\n--- Single Rational Resampler Calculation ---\n");
    printf("Target: Input Rate = %.2f Hz, Output Rate = %.2f Hz\n", inputs->input_rate, inputs->output_rate);
    double target_ratio = (inputs->input_rate != 0) ? inputs->output_rate / inputs->input_rate : NAN;
    if (isfinite(target_ratio)) { printf("Target Ratio (output/input) = %.10f\n", target_ratio); }
    else { printf("Target Ratio (output/input) = Invalid\n"); }
    printf("Match Tolerance (relative error): %.1e\n", MATCH_TOLERANCE);
    printf("Arbitrary Suggestion Threshold (relative error): %.1e\n", ARBITRARY_SUGGEST_ERROR_THRESHOLD);
    printf("CFE searching with max denominator M <= %lld\n", HARD_MAX_LM_LIMIT);
    fflush(stdout); // Ensure context is printed before potentially long calculation

    // --- Perform Calculation ---
    long long l, m;
    double achieved_rate;
    StatusCode status = calculate_single_resampler(inputs->input_rate, inputs->output_rate, &l, &m, &achieved_rate);

    // --- Print Results or Failure Message ---
    if (status == SUCCESS) {
        print_single_resampler_result(inputs->input_rate, inputs->output_rate, l, m, achieved_rate);
    } else {
        // Calculation function or CFE utils should have printed specific errors.
        // Add a generic failure message here. (Message unchanged)
        printf("\nCalculation failed. No results to display for single resampler.\n");
        fflush(stdout);
    }
    return status; // Return the status from the calculation
}
