// src/two_resamplers.c

#include "two_resamplers.h" // Include own header first
#include "signal_handler.h" // For g_shutdown_requested
#include "utils.h"          // For print_stderr, status codes, status_code_to_string, max_long_long
#include "resampling_common.h"// For TwoResamplerResult, constants, thresholds
#include "cfe_utils.h"      // For CFE/Factorization utilities
#include <stdio.h>
#include <stdlib.h> // For malloc, free, memset
#include <math.h>   // For fabs, isfinite, NAN, INFINITY, sqrt, fmax, HUGE_VAL, DBL_MAX, DBL_MIN, copysign
#include <float.h>  // For DBL_MAX, DBL_MIN
#include <string.h> // For memset, strerror (used indirectly)
#include <errno.h>  // For errno (used indirectly)
#include <limits.h> // For LLONG_MAX, LLONG_MIN, INT_MAX
#include <stdint.h> // For ULLONG_MAX if needed
#include <stdbool.h> // For bool type

#define DOT_FREQUENCY_CFE 10    // Frequency for printing progress dots (now applies to convergent loop)
#define DOT_FREQUENCY_GH 1000   // Frequency for printing dots during g/h loop (can be very large)
#define ERROR_EPSILON 1e-15     // Small tolerance for comparing floating point errors

// --- Static Helper Prototypes ---
static double calculate_achieved_output_rate_two(double input_rate, long long l1, long long m1, long long l2, long long m2);
static void print_two_resamplers_result(const char* result_title, const UserInputs* inputs, const TwoResamplerResult* result);
static void assess_candidate_and_update(long long l1, long long m1, long long l2, long long m2,
                                        long long max_lm_per_stage,
                                        const UserInputs* inputs,
                                        SearchMode comparison_mode,
                                        TwoResamplerResult* final_result,
                                        bool* first_dual_match_found_in_default_ptr); // Optional optimization flag
static const char* priority_to_string(IntermediatePriority priority);
// Removed max_ll prototype - using consolidated version from utils.h

// --- Static Helper Implementations ---

/**
 * @brief Maps IntermediatePriority enum to a descriptive string.
 */
static const char* priority_to_string(IntermediatePriority priority)
{
    switch (priority)
    {
        case PRIORITIZE_STRICT_INTERMEDIATE: return "Strict Intermediate";
        case PRIORITIZE_FINAL_RATE:          return "Final Rate";
        case PRIORITIZE_BALANCED_DUAL:       return "Balanced Dual";
        case PRIORITIZE_BEST_OVERALL:        return "Best Overall (Ignoring Intermediate)";
        default:                             return "Unknown";
    }
}

/**
 * @brief Calculates the final output rate for a two-stage resampler.
 * @details Uses double precision internally to minimize overflow risk.
 * @return Achieved rate, or NAN on error (invalid inputs, overflow).
 */
static double calculate_achieved_output_rate_two(double input_rate, long long l1, long long m1, long long l2, long long m2)
{
    // Basic validation: L/M must be positive, input rate positive and finite
    if (m1 <= 0 || m2 <= 0 || l1 <= 0 || l2 <= 0 || !isfinite(input_rate) || input_rate <= 0) return NAN;

    // Calculate intermediate ratios using double precision
    double r1_dbl = (double)l1 / (double)m1;
    double r2_dbl = (double)l2 / (double)m2;

    // Check if intermediate ratios are valid
    if (!isfinite(r1_dbl) || !isfinite(r2_dbl)) return NAN;

    // Calculate combined ratio, checking for potential overflow/underflow
    // Check overflow: |r1*r2| > DBL_MAX
    // Check underflow: |r1*r2| < DBL_MIN (Treat significant underflow as error for simplicity)
    if ((fabs(r1_dbl) > 1.0 && fabs(r2_dbl) > 0.0 && fabs(r1_dbl) > DBL_MAX / fabs(r2_dbl)) || // Overflow
        (fabs(r1_dbl) > 0.0 && fabs(r2_dbl) > 0.0 && fabs(r1_dbl) < DBL_MIN / fabs(r2_dbl))) { // Underflow towards 0
        return NAN;
    }
    double combined_ratio = r1_dbl * r2_dbl;
    if (!isfinite(combined_ratio)) return NAN;

    // Calculate final rate, checking for potential overflow/underflow
    // Check overflow: |input*ratio| > DBL_MAX
    // Check underflow: |input*ratio| < DBL_MIN
    if ((fabs(input_rate) > 1.0 && fabs(combined_ratio) > 0.0 && fabs(input_rate) > DBL_MAX / fabs(combined_ratio)) || // Overflow
        (fabs(input_rate) > 0.0 && fabs(combined_ratio) > 0.0 && fabs(input_rate) < DBL_MIN / fabs(combined_ratio))) { // Underflow towards 0
        return NAN;
    }
    double result = input_rate * combined_ratio;

    return isfinite(result) ? result : NAN; // Return NAN if final result is not finite
}


/**
 * @brief Prints the results of the two-stage resampler search.
 * @details Format preserved. Calculates display values based on the best result found.
 *          Includes L/M values, complexity, rates, errors (abs/rel/ppm), and notes.
 */
static void print_two_resamplers_result(const char* result_title, // e.g., "Minimal Sum Search Results:"
                                        const UserInputs* inputs,
                                        const TwoResamplerResult* result)
{
    if (g_shutdown_requested) return;
    if (!result || !inputs || !result_title)
    {
        fprintf(stderr, "Internal Error (print_two): NULL pointer provided.\n");
        return;
    }

    // --- Print Merged Title and Status Line ---
    if (result->result_found)
    {
        // Determine status based on error and factor size
        bool is_match = result->match_found;
        double relative_difference_for_status = NAN;
         if (isfinite(result->min_error) && fabs(inputs->output_rate) > CFE_DBL_EPSILON) {
             relative_difference_for_status = fabs(result->min_error / inputs->output_rate);
         } else if (isfinite(result->min_error)) {
             relative_difference_for_status = fabs(result->min_error); // Use abs error if target near zero
         }

        bool significant_error = !is_match && isfinite(relative_difference_for_status) && (relative_difference_for_status >= ARBITRARY_SUGGEST_ERROR_THRESHOLD);
        bool large_factors = (result->best_l1 > SUGGEST_ARBITRARY_LM_THRESHOLD || result->best_m1 > SUGGEST_ARBITRARY_LM_THRESHOLD ||
                              result->best_l2 > SUGGEST_ARBITRARY_LM_THRESHOLD || result->best_m2 > SUGGEST_ARBITRARY_LM_THRESHOLD);

        // (Prompts unchanged)
        printf("%s ", result_title);
        if (is_match && large_factors) {
            printf("Match Found with Large L/M\n");
        } else if (is_match) {
            printf("Match Found!\n");
        } else if (significant_error) {
            printf("Caution! Significant Error\n");
        } else {
            printf("Approximation Found\n");
        }

        // --- Print Result Details ---
        // Calculate display values from the stored best result
        double achieved_output_disp = calculate_achieved_output_rate_two(inputs->input_rate, result->best_l1, result->best_m1, result->best_l2, result->best_m2);
        double rate_difference_disp = isfinite(achieved_output_disp) ? achieved_output_disp - inputs->output_rate : NAN;
        double relative_difference_disp = NAN; // Absolute relative difference for display
        double overall_ppm = NAN; // Signed PPM error
        if (isfinite(rate_difference_disp)) {
             if (fabs(inputs->output_rate) > CFE_DBL_EPSILON) {
                 overall_ppm = (rate_difference_disp / inputs->output_rate) * 1000000.0;
                 relative_difference_disp = fabs(rate_difference_disp / inputs->output_rate);
             } else { // Target is near zero
                 relative_difference_disp = fabs(rate_difference_disp);
                 overall_ppm = (fabs(rate_difference_disp) > CFE_DBL_EPSILON) ? copysign(INFINITY, rate_difference_disp) : 0.0;
             }
        }

        double inter_rate_disp = result->achieved_intermediate_rate; // Already stored
        double inter_diff_disp = NAN;        // Signed intermediate difference
        double inter_rel_diff_disp = NAN;    // Absolute relative intermediate difference
        double intermediate_ppm = NAN;       // Signed intermediate PPM error
        if (inputs->intermediate_rate_targeted && isfinite(inter_rate_disp)) {
            inter_diff_disp = inter_rate_disp - inputs->target_intermediate_rate;
            if (fabs(inputs->target_intermediate_rate) > CFE_DBL_EPSILON) {
                intermediate_ppm = (inter_diff_disp / inputs->target_intermediate_rate) * 1000000.0;
                inter_rel_diff_disp = fabs(inter_diff_disp / inputs->target_intermediate_rate);
            } else { // Target intermediate rate is near zero
                 inter_rel_diff_disp = fabs(inter_diff_disp);
                 intermediate_ppm = (fabs(inter_diff_disp) > CFE_DBL_EPSILON) ? copysign(INFINITY, inter_diff_disp) : 0.0;
            }
        }

        // Use consolidated helper for stage complexity
        long long stage1_max = max_long_long(result->best_l1, result->best_m1);
        long long stage2_max = max_long_long(result->best_l2, result->best_m2);
        long long sum_of_maxima = -1; // Overall complexity metric
        if (stage1_max >= 0 && stage2_max >= 0) {
             // Check for overflow when adding stage maxima
             if (LLONG_MAX - stage2_max < stage1_max) { sum_of_maxima = LLONG_MAX; } // Indicate overflow (Corrected check)
             else { sum_of_maxima = stage1_max + stage2_max; }
        }


        // L/M values - (Prompts unchanged)
        printf("\nFirst Resampler: Complexity: %lld\n", stage1_max);
        printf("  L1 = %lld (Interpolation)\n", result->best_l1);
        printf("  M1 = %lld (Decimation)\n", result->best_m1);

        printf("\nSecond Resampler: Complexity: %lld\n", stage2_max);
        printf("  L2 = %lld (Interpolation)\n", result->best_l2);
        printf("  M2 = %lld (Decimation)\n", result->best_m2);

        // Sums/Rates Output - (Prompts unchanged)
        printf("\nOverall Sum (L1+M1+L2+M2) = %lld\n", result->sum_coeffs); // Sum checked during search
        printf("Overall Complexity (Sum of Max) = %lld\n", sum_of_maxima);
        printf("Achieved Combined Ratio = (%lld / %lld) * (%lld / %lld) = %.10f\n",
               result->best_l1, result->best_m1, result->best_l2, result->best_m2, result->best_combined_ratio);
        printf("Achieved Intermediate Rate = ");
        if (isfinite(inter_rate_disp)) printf("%.10f Hz\n", inter_rate_disp); else printf("Calculation Error\n");
        printf("Achieved Output Rate = ");
        if (isfinite(achieved_output_disp)) printf("%.10f Hz\n", achieved_output_disp); else printf("Calculation Error\n");

        // Difference and PPM Details - (Prompts unchanged)
        printf("Overall Rate Difference: ");
        if (isfinite(rate_difference_disp)) printf("%+.10f Hz, Relative: %.2e\n", rate_difference_disp, relative_difference_disp);
        else printf("Calculation Error\n");
        printf("Overall PPM Error: ");
        if (isfinite(overall_ppm)) printf("%+.3f PPM\n", overall_ppm);
        else if (isnan(overall_ppm)) printf("N/A\n");
        else printf("%sInfinity\n", (overall_ppm > 0) ? "+" : "-");


        // Intermediate Difference (Conditional) - (Prompts unchanged)
        if (inputs->intermediate_rate_targeted) {
            printf("Intermediate Rate Difference: ");
            if (isfinite(inter_diff_disp)) printf("%+.10f Hz, Relative: %.2e\n", inter_diff_disp, inter_rel_diff_disp);
            else printf("Calculation Error\n");
            printf("Intermediate PPM Error: ");
             if (isfinite(intermediate_ppm)) printf("%+.3f PPM\n", intermediate_ppm);
             else if (isnan(intermediate_ppm)) printf("N/A\n");
             else printf("%sInfinity\n", (intermediate_ppm > 0) ? "+" : "-");
        }


        // --- Notes Section ---
        // (Logic and Prompts unchanged)
        bool inter_mismatch_notice = inputs->intermediate_rate_targeted && !result->intermediate_match_found;
        if (significant_error || large_factors || (!is_match && !significant_error) || inter_mismatch_notice)
        {
             printf("\n"); // Spacing before notes block
             if (significant_error) {
                 printf("Caution!: The best approximation found differs significantly\n");
                 printf("          from the target output rate (Final Relative Error: %.2e).\n", relative_difference_disp);
                 printf("          An arbitrary (non-rational) resampling method might be required\n");
                 printf("          for higher accuracy if this error is unacceptable.");
             }
             if (large_factors) {
                 if (significant_error) printf("\n");
                 printf("Note: One or more calculated factors (L1=%lld, M1=%lld, L2=%lld, M2=%lld) are large.\n", result->best_l1, result->best_m1, result->best_l2, result->best_m2);
                 printf("      Implementing this rational ratio with standard polyphase filters may be\n");
                 printf("      computationally expensive or require long filters.\n");
                 printf("      Consider if an arbitrary resampling method might be a suitable\n");
                 printf("      alternative if filter complexity is a primary concern.\n");
             }
             else if (!is_match && !significant_error) {
                 printf("Note: Achieved final rate differs slightly from target rate (Rel Diff: %.2e).\n", relative_difference_disp);
             }

             if (inter_mismatch_notice) {
                 if (significant_error || large_factors || (!is_match && !significant_error)) printf("\n");
                 printf("Note: Achieved intermediate rate differs from target rate");
                 if (isfinite(inter_rel_diff_disp)) printf(" (Rel Diff: %.2e).\n", inter_rel_diff_disp); else printf(".\n");

                 if (inputs->intermediate_priority == PRIORITIZE_STRICT_INTERMEDIATE) {
                     printf("      (Strict intermediate target was requested but could not be met.)\n");
                 }
             }
        }
         // --- End Notes Section ---

    }
    else // result_found is false
    {
         // (Prompts unchanged)
         printf("%s Not found\n", result_title);
         printf("\nPossible reasons:\n");
         if (inputs->max_lm_per_stage > 0 && inputs->search_mode == SEARCH_DEFAULT) printf("  - Maximum L/M value (%lld) is too small.\n", inputs->max_lm_per_stage);
         else if (inputs->search_mode == SEARCH_MINIMAL_SUM) printf("  - Internal L/M limit (%lld) might be too small.\n", HARD_MAX_LM_LIMIT);
         else printf("  - Factorization or internal limits might be the issue.\n");
         printf("  - Required ratio is difficult to approximate.\n");
         if (inputs->intermediate_rate_targeted) { printf("  - Could not meet intermediate rate target with priority '%s'.\n", priority_to_string(inputs->intermediate_priority));}
         if (!inputs->allow_m1_greater_than_l1) { printf("  - Required L1>=M1 constraint is too restrictive for this ratio.\n"); }
         printf("Check constraints, priority, and consider if a larger maximum L/M is needed.\n");
    }
    fflush(stdout);
}


/**
 * @brief Evaluates a candidate two-stage configuration (L1, M1, L2, M2) and updates the best result if superior.
 *
 * @details
 * This function is the core comparison logic for the two-stage search. It takes a single
 * candidate set of factors and checks if it's valid (within limits, positive, satisfies L1>=M1 if required).
 * It then calculates the achieved overall and intermediate rates and their errors.
 * Based on the user's chosen SearchMode (Minimal Sum vs Default/Custom Limit) and
 * IntermediatePriority, it compares the candidate's metrics (error, sum of coefficients)
 * against the current best result stored in `final_result`.
 * If the candidate is deemed better according to the active rules, `final_result` is updated.
 *
 * @param l1 Candidate L1 factor.
 * @param m1 Candidate M1 factor.
 * @param l2 Candidate L2 factor.
 * @param m2 Candidate M2 factor.
 * @param max_lm_per_stage The maximum allowed value for any single L or M factor.
 * @param inputs Pointer to the UserInputs struct containing constraints and priorities.
 * @param comparison_mode The active search mode (SEARCH_MINIMAL_SUM or SEARCH_DEFAULT).
 * @param final_result Pointer to the TwoResamplerResult struct holding the current best result (will be updated).
 * @param first_dual_match_found_in_default_ptr Optional optimization flag (can be NULL). If non-NULL and
 *        in DEFAULT mode with BALANCED_DUAL priority, this flag is set to true when the first
 *        candidate meeting *both* rate targets is found, allowing potential early exit.
 */
static void assess_candidate_and_update(long long l1, long long m1, long long l2, long long m2,
                                        long long max_lm_per_stage,
                                        const UserInputs* inputs,
                                        SearchMode comparison_mode,
                                        // double target_ratio, // Not strictly needed here, errors calculated vs rates
                                        TwoResamplerResult* final_result,
                                        bool* first_dual_match_found_in_default_ptr)
{
    // --- Basic validation of the candidate factors ---
    if (l1 <= 0 || m1 <= 0 || l2 <= 0 || m2 <= 0 ||           // Factors must be positive
        l1 > max_lm_per_stage || m1 > max_lm_per_stage ||     // Check against per-stage limit
        l2 > max_lm_per_stage || m2 > max_lm_per_stage)
    {
        return; // Invalid factors or exceeds per-stage limit
    }
    // Check L1 >= M1 constraint if active
    if (!inputs->allow_m1_greater_than_l1 && l1 < m1) {
        return; // Violates constraint
    }

    // --- Calculate metrics for the current candidate ---

    // Achieved output rate and absolute error vs target output rate
    double current_overall_error = INFINITY;
    bool current_overall_match_ok = false; // Does it meet final rate MATCH_TOLERANCE?
    double achieved_output = calculate_achieved_output_rate_two(inputs->input_rate, l1, m1, l2, m2);
    if (isfinite(achieved_output)) {
        current_overall_error = fabs(achieved_output - inputs->output_rate); // Store absolute error
        // Check relative error against overall MATCH_TOLERANCE
        double rel_diff = (fabs(inputs->output_rate) > CFE_DBL_EPSILON)
                          ? (current_overall_error / fabs(inputs->output_rate))
                          : current_overall_error; // Use absolute if target near zero
        current_overall_match_ok = (rel_diff < MATCH_TOLERANCE);
    }
    // If achieved_output is not finite, error remains INFINITY, match remains false.

    // Achieved intermediate rate and absolute error vs target intermediate rate (if applicable)
    double current_intermediate_rate = NAN;
    double current_intermediate_error = INFINITY;
    bool current_intermediate_match_ok = false; // Meets INTERMEDIATE_MATCH_TOLERANCE?
    // Calculate only if targeted OR needed for comparison logic later
    if (inputs->intermediate_rate_targeted || inputs->intermediate_priority != PRIORITIZE_BEST_OVERALL) {
         // Need input_rate * (l1/m1)
         if (m1 > 0 && isfinite(inputs->input_rate) && inputs->input_rate > 0) {
             // Use double for intermediate calculation to avoid premature overflow
             double inter_rate_dbl = inputs->input_rate * ((double)l1 / (double)m1);
             // Check for potential overflow/underflow in intermediate rate calculation itself
             if ((fabs(inputs->input_rate) > 1.0 && fabs((double)l1 / (double)m1) > 0.0 && fabs(inputs->input_rate) > DBL_MAX / fabs((double)l1 / (double)m1)) || // Overflow
                 (fabs(inputs->input_rate) > 0.0 && fabs((double)l1 / (double)m1) > 0.0 && fabs(inputs->input_rate) < DBL_MIN / fabs((double)l1 / (double)m1))) { // Underflow
                 current_intermediate_rate = NAN;
             } else {
                 current_intermediate_rate = isfinite(inter_rate_dbl) ? inter_rate_dbl : NAN;
             }


             if (inputs->intermediate_rate_targeted && isfinite(current_intermediate_rate)) {
                 current_intermediate_error = fabs(current_intermediate_rate - inputs->target_intermediate_rate);
                 // Check relative intermediate error against tolerance
                 double rel_inter_diff = (fabs(inputs->target_intermediate_rate) > CFE_DBL_EPSILON)
                                         ? (current_intermediate_error / fabs(inputs->target_intermediate_rate))
                                         : current_intermediate_error;
                 current_intermediate_match_ok = (rel_inter_diff < INTERMEDIATE_MATCH_TOLERANCE);
             }
             // If not targeted but calculated, error remains INFINITY, match remains false.
         }
         // If cannot calculate intermediate rate (input invalid, m1=0), rate is NAN, error INFINITY.
    }
    // If intermediate priority is BEST_OVERALL and not targeted, we don't need intermediate error for comparison.
    // Calculate the rate anyway if possible, just for storage in result if this candidate wins.
    else if (isfinite(inputs->input_rate) && inputs->input_rate > 0 && m1 > 0) {
         double inter_rate_dbl = inputs->input_rate * ((double)l1 / (double)m1);
         // Check for potential overflow/underflow
         if ((fabs(inputs->input_rate) > 1.0 && fabs((double)l1 / (double)m1) > 0.0 && fabs(inputs->input_rate) > DBL_MAX / fabs((double)l1 / (double)m1)) || // Overflow
             (fabs(inputs->input_rate) > 0.0 && fabs((double)l1 / (double)m1) > 0.0 && fabs(inputs->input_rate) < DBL_MIN / fabs((double)l1 / (double)m1))) { // Underflow
             current_intermediate_rate = NAN;
         } else {
             current_intermediate_rate = isfinite(inter_rate_dbl) ? inter_rate_dbl : NAN;
         }
         // Error remains INFINITY as it's not used for selection in this case.
    }


    // Calculate sum L1+M1+L2+M2 safely using unsigned long long for intermediate steps
    long long current_sum = -1; // Use -1 to indicate overflow or invalid
    unsigned long long temp_sum_ull = 0;
    // Check overflow at each addition stage
    if ((unsigned long long)l1 > ULLONG_MAX - (unsigned long long)m1 ||
        (temp_sum_ull = (unsigned long long)l1 + (unsigned long long)m1) > ULLONG_MAX - (unsigned long long)l2 ||
        (temp_sum_ull += (unsigned long long)l2) > ULLONG_MAX - (unsigned long long)m2)
    {
        current_sum = -1; // Overflow detected during intermediate sum
    } else {
        temp_sum_ull += (unsigned long long)m2; // Final addition
        // Check if final unsigned sum fits in signed long long
        current_sum = (temp_sum_ull <= (unsigned long long)LLONG_MAX) ? (long long)temp_sum_ull : -1;
    }

    // --- Get previous best result properties for comparison ---
    bool previous_result_exists = final_result->result_found;
    bool previous_overall_match_ok = previous_result_exists && final_result->match_found;
    bool previous_intermediate_match_ok = previous_result_exists && final_result->intermediate_match_found;
    double previous_overall_error = previous_result_exists ? final_result->min_error : INFINITY;
    double previous_intermediate_error = previous_result_exists ? final_result->intermediate_rate_error : INFINITY;
    long long previous_sum = previous_result_exists ? final_result->sum_coeffs : LLONG_MAX;
    if (previous_sum < 0) previous_sum = LLONG_MAX; // Treat potential stored -1 as max for comparison


    // --- Comparison Logic: Determine if the current candidate is better than the previous best ---
    bool should_update = false; // Assume current is not better initially

    // 1. Handle invalid current candidate (overflow in sum or rate calculation error)
    if (current_sum == -1 || !isfinite(current_overall_error))
    {
        should_update = false; // Cannot use this candidate
    }
    // 2. Handle the very first valid candidate found
    else if (!previous_result_exists)
    {
        // If strict intermediate priority is set, the first candidate *must* meet it.
        if (inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_STRICT_INTERMEDIATE) {
             should_update = current_intermediate_match_ok; // Only accept if strict target met
        } else {
             should_update = true; // Any valid result is acceptable as the first one otherwise
        }
    }
    // 3. Compare current valid candidate against existing best result
    else
    {
        // Determine the effective priority (defaults to BEST_OVERALL if not targeted)
        IntermediatePriority effective_priority = inputs->intermediate_rate_targeted
                                                  ? inputs->intermediate_priority
                                                  : PRIORITIZE_BEST_OVERALL;
        bool current_is_better = false; // Flag specific to this comparison block

        switch (effective_priority)
        {
            // --- Case 1: Strict Intermediate Priority ---
            case PRIORITIZE_STRICT_INTERMEDIATE:
                // Candidate must meet intermediate target to even be considered.
                if (current_intermediate_match_ok) {
                    // If previous best didn't meet intermediate target, current is better.
                    if (!previous_intermediate_match_ok) {
                        current_is_better = true;
                    }
                    // If both meet intermediate target, compare based on the primary search mode (Sum or Default Error/Sum).
                    else {
                        if (comparison_mode == SEARCH_MINIMAL_SUM) {
                            // Lower sum wins. Tie-break with lower overall error.
                            if (current_sum < previous_sum) current_is_better = true;
                            else if (current_sum == previous_sum && current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                        } else { // SEARCH_DEFAULT
                            // Lower overall error wins. Tie-break with lower sum.
                            if (current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                            else if (fabs(current_overall_error - previous_overall_error) < ERROR_EPSILON && current_sum < previous_sum) current_is_better = true;
                        }
                    }
                }
                // If current doesn't meet intermediate target, it cannot be better under strict priority.
                break;

            // --- Case 2: Prioritize Final Rate, then Intermediate ---
            case PRIORITIZE_FINAL_RATE:
                // First, check if final rate match status improved.
                // If current matches final rate and previous didn't, current is better.
                if (current_overall_match_ok && !previous_overall_match_ok) { current_is_better = true; }
                // If current doesn't match and previous did, current is worse.
                else if (!current_overall_match_ok && previous_overall_match_ok) { current_is_better = false; }
                // If both match or both don't match final rate, compare based on intermediate closeness (if targeted).
                else {
                    // If intermediate targeted: Lower intermediate error wins.
                    if (inputs->intermediate_rate_targeted && current_intermediate_error < previous_intermediate_error - ERROR_EPSILON) { current_is_better = true; }
                    else if (inputs->intermediate_rate_targeted && previous_intermediate_error < current_intermediate_error - ERROR_EPSILON) { current_is_better = false; }
                    // If intermediate errors are equal or not targeted, fall back to primary search mode comparison.
                    else {
                        if (comparison_mode == SEARCH_MINIMAL_SUM) {
                            if (current_sum < previous_sum) current_is_better = true;
                            else if (current_sum == previous_sum && current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                        } else { // SEARCH_DEFAULT
                            if (current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                            else if (fabs(current_overall_error - previous_overall_error) < ERROR_EPSILON && current_sum < previous_sum) current_is_better = true;
                        }
                    }
                }
                break;

            // --- Case 3: Balanced Dual Priority ---
            case PRIORITIZE_BALANCED_DUAL:
            { // Use block scope for local variables
                // Check if candidate meets both intermediate (if targeted) and final rate targets.
                // Only consider dual match if intermediate rate *was actually targeted*.
                bool current_is_dual = inputs->intermediate_rate_targeted && current_overall_match_ok && current_intermediate_match_ok;
                bool previous_is_dual = inputs->intermediate_rate_targeted && previous_overall_match_ok && previous_intermediate_match_ok;

                // If current achieves dual match and previous didn't, current is better.
                if (current_is_dual && !previous_is_dual) { current_is_better = true; }
                // If previous was dual and current isn't, current is worse.
                else if (!current_is_dual && previous_is_dual) { current_is_better = false; }
                // If both are dual matches OR neither is (or intermediate not targeted)
                else {
                    // If both are dual matches, compare using primary search mode criteria.
                    if (current_is_dual) { // Implies previous_is_dual is also true, and intermediate was targeted
                         if (comparison_mode == SEARCH_MINIMAL_SUM) {
                             if (current_sum < previous_sum) current_is_better = true;
                             else if (current_sum == previous_sum && current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                         } else { // SEARCH_DEFAULT
                             if (current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                             else if (fabs(current_overall_error - previous_overall_error) < ERROR_EPSILON && current_sum < previous_sum) current_is_better = true;
                         }
                    }
                    // If neither is a dual match (or not targeted), fall back to PRIORITIZE_FINAL_RATE logic.
                    else {
                        if (current_overall_match_ok && !previous_overall_match_ok) { current_is_better = true; }
                        else if (!current_overall_match_ok && previous_overall_match_ok) { current_is_better = false; }
                        else { // Same final rate status
                            // If intermediate was targeted, compare intermediate errors
                            if (inputs->intermediate_rate_targeted && current_intermediate_error < previous_intermediate_error - ERROR_EPSILON) { current_is_better = true; }
                            else if (inputs->intermediate_rate_targeted && previous_intermediate_error < current_intermediate_error - ERROR_EPSILON) { current_is_better = false; }
                            else { // Intermediate errors equal or not targeted, use search mode
                                if (comparison_mode == SEARCH_MINIMAL_SUM) {
                                    if (current_sum < previous_sum) current_is_better = true;
                                    else if (current_sum == previous_sum && current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                                } else { // SEARCH_DEFAULT
                                    if (current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                                    else if (fabs(current_overall_error - previous_overall_error) < ERROR_EPSILON && current_sum < previous_sum) current_is_better = true;
                                }
                            }
                        }
                    } // end fallback
                } // end both dual or neither dual case
            } // End of case block scope
            break;

            // --- Case 4: Best Overall (Ignore Intermediate for Selection) ---
            case PRIORITIZE_BEST_OVERALL:
                 // If current matches final rate and previous didn't, current is better.
                 if (current_overall_match_ok && !previous_overall_match_ok) { current_is_better = true; }
                 // If current doesn't match and previous did, current is worse.
                 else if (!current_overall_match_ok && previous_overall_match_ok) { current_is_better = false; }
                 // If both match or both don't match, compare using primary search mode criteria only.
                 else {
                     if (comparison_mode == SEARCH_MINIMAL_SUM) {
                         if (current_sum < previous_sum) current_is_better = true;
                         else if (current_sum == previous_sum && current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                     } else { // SEARCH_DEFAULT
                         if (current_overall_error < previous_overall_error - ERROR_EPSILON) current_is_better = true;
                         else if (fabs(current_overall_error - previous_overall_error) < ERROR_EPSILON && current_sum < previous_sum) current_is_better = true;
                     }
                 }
                 break;
        } // End switch (effective_priority)

        should_update = current_is_better; // Set final decision based on comparison logic
    } // End else (previous result exists)


    // --- Update final result if current candidate is deemed better ---
    if (should_update) {
        final_result->best_l1 = l1;
        final_result->best_m1 = m1;
        final_result->best_l2 = l2;
        final_result->best_m2 = m2;
        // Calculate best combined ratio only when updating (avoid redundant calc)
        double r1_upd = (m1 != 0) ? (double)l1 / (double)m1 : NAN;
        double r2_upd = (m2 != 0) ? (double)l2 / (double)m2 : NAN;
        if (isfinite(r1_upd) && isfinite(r2_upd)) {
             // Check overflow before final multiplication
             if ((fabs(r1_upd) > 1.0 && fabs(r2_upd) > 0.0 && fabs(r1_upd) > DBL_MAX / fabs(r2_upd)) ||
                 (fabs(r1_upd) > 0.0 && fabs(r2_upd) > 0.0 && fabs(r1_upd) < DBL_MIN / fabs(r2_upd))) {
                  final_result->best_combined_ratio = NAN;
             } else {
                 final_result->best_combined_ratio = r1_upd * r2_upd;
                 if (!isfinite(final_result->best_combined_ratio)) final_result->best_combined_ratio = NAN;
             }
        } else {
            final_result->best_combined_ratio = NAN;
        }

        final_result->min_error = current_overall_error; // Store absolute overall error
        final_result->result_found = true;               // Mark that at least one valid result exists
        final_result->match_found = current_overall_match_ok; // Store overall match status
        final_result->sum_coeffs = current_sum;          // Store sum of coefficients
        final_result->achieved_intermediate_rate = current_intermediate_rate; // Store achieved intermediate rate
        final_result->intermediate_rate_error = current_intermediate_error; // Store absolute intermediate error
        final_result->intermediate_match_found = current_intermediate_match_ok; // Store intermediate match status

        // Optimization: Signal if a dual match was found in Default+Balanced mode
        if (first_dual_match_found_in_default_ptr &&
            comparison_mode == SEARCH_DEFAULT &&
            inputs->intermediate_rate_targeted && // Check if actually targeted
            inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL &&
            current_overall_match_ok && current_intermediate_match_ok)
        {
             *first_dual_match_found_in_default_ptr = true;
        }
    }
} // ** END of assess_candidate_and_update **

// --- Core Search Function ---

/**
 * @brief Performs the core two-stage resampler search using CFE and factorization.
 *
 * @details
 * 1. Calculates CFE convergents (L/M) for the target ratio up to derived max denominator.
 * 2. Finds the simplest exact ratio Ls/Ms from the CFE convergents.
 * 3. Iterates through CFE convergents L/M:
 *    a. Factors L into pairs (l1, l2).
 *    b. Factors M into pairs (m1, m2).
 *    c. Calls `assess_candidate_and_update` for permutations like (l1, m1, l2, m2), (l1, m2, l2, m1), etc.
 * 4. If an exact Ls/Ms was found, iterates through its factor pairs (ls1,ls2), (ms1,ms2)
 *    and considers multiples by common factors g, h: (g*ls1, h*ms1, h*ls2, g*ms2).
 *    Calculates maximum g and h based on max_lm_per_stage.
 *    Checks for overflow when calculating L1,M1,L2,M2.
 *    Calls `assess_candidate_and_update` for these generated candidates.
 *    **Optimization:** In Minimal Sum mode, the g/h search stops after checking g=1, h=1.
 * Checks for shutdown requests frequently. Populates the `final_result` struct.
 *
 * @param inputs Pointer to UserInputs containing configuration.
 * @param max_lm_per_stage The effective max L/M value per stage for this search.
 * @param search_mode The active search mode (Minimal Sum or Default).
 * @param final_result Pointer to the TwoResamplerResult struct to store the best result found.
 * @param conv_count_out Pointer to store the number of CFE convergents processed (for info messages).
 *
 * @return SUCCESS if a valid result was found and stored in final_result,
 *         FAILURE if no suitable result was found or shutdown requested,
 *         Specific error codes (ERR_*) on critical failures.
 */
static StatusCode perform_two_stage_search(const UserInputs* inputs,
                                           long long max_lm_per_stage, SearchMode search_mode,
                                           TwoResamplerResult* final_result,
                                           int* conv_count_out)
{
    // --- Initializations ---
    if (g_shutdown_requested) return FAILURE;
    if (!inputs || !final_result || !conv_count_out) return ERR_INVALID_INPUT;
    fflush(stdout); // Ensure prompts are visible before potentially long search

    // Initialize result struct thoroughly
    memset(final_result, 0, sizeof(TwoResamplerResult));
    final_result->min_error = INFINITY; final_result->match_found = false;
    final_result->result_found = false; final_result->sum_coeffs = LLONG_MAX;
    final_result->best_combined_ratio = NAN; final_result->achieved_intermediate_rate = NAN;
    final_result->intermediate_rate_error = INFINITY; final_result->intermediate_match_found = false;
    final_result->best_l1 = 0; final_result->best_m1 = 0;
    final_result->best_l2 = 0; final_result->best_m2 = 0;

    // State variables
    bool first_dual_match_found_in_default = false; // Optimization flag
    *conv_count_out = -1; // Init convergents processed count
    StatusCode status = SUCCESS;
    int dots_printed_cfe = 0; // Counter for progress dots (CFE loop)
    int dots_printed_gh = 0;  // Counter for progress dots (g/h loop)
    unsigned long long gh_iterations = 0; // Counter for g/h iterations for dot printing
    bool progress_msg_cfe_printed = false; // Flag for CFE progress message timing
    bool progress_msg_gh_printed = false;  // Flag for g/h progress message timing

    // Dynamic arrays (potentially reused, declare outside loops)
    Convergent* convergents = NULL; int conv_count = 0;
    FactorPair* l_factors = NULL; int l_factor_count = 0;
    FactorPair* m_factors = NULL; int m_factor_count = 0;

    // Simplest Exact Ratio variables
    bool simplest_exact_found = false;
    long long L_simplest = 0; long long M_simplest = 0;
    const double EXACT_RATIO_EPSILON = 1e-14; // Tolerance for finding simplest exact L/M

    // --- Calculate Target Ratio & Max Overall Denominator for CFE ---
    if (inputs->input_rate <= 0) { print_stderr("Input rate must be positive."); status = ERR_INVALID_INPUT; goto cleanup_search; }
    double target_ratio = inputs->output_rate / inputs->input_rate;
    if (!isfinite(target_ratio) || target_ratio <= 0) { print_stderr("Cannot calculate valid positive target ratio."); status = ERR_CALCULATION; goto cleanup_search; }

    // Determine max denominator for CFE based on max_lm_per_stage
    // max_overall_Q = max_M = max(M1*M2) <= max_lm_per_stage * max_lm_per_stage
    long long max_overall_denominator = LLONG_MAX;
    if (max_lm_per_stage > 0) {
         // Check for overflow before squaring max_lm_per_stage
         unsigned long long max_lm_ull = (unsigned long long)max_lm_per_stage;
         const unsigned long long SQRT_LLONG_MAX_APPROX = 3037000499ULL; // floor(sqrt(LLONG_MAX))
         if (max_lm_ull > SQRT_LLONG_MAX_APPROX) {
             max_overall_denominator = LLONG_MAX; // Squaring would overflow signed long long
         } else {
             unsigned long long squared = max_lm_ull * max_lm_ull;
             // Assign if squared value fits in signed long long
             max_overall_denominator = (squared <= (unsigned long long)LLONG_MAX) ? (long long)squared : LLONG_MAX;
         }
    }
    // Ensure max_overall_denominator is at least 1
    if (max_overall_denominator <= 0 && max_overall_denominator != LLONG_MAX) {
         max_overall_denominator = 1;
    }


    // --- Step 1: Run CFE ---
    status = calculate_cfe_convergents(target_ratio, max_overall_denominator, &convergents, &conv_count);
    *conv_count_out = conv_count; // Store count for informational messages
    if (status != SUCCESS) {
        if (status < FAILURE) { print_stderr("Critical Error during CFE: %s.", status_code_to_string(status)); }
        else if (conv_count == 0) { print_stderr("Info: CFE generated no convergents (check max L/M and ratio)."); }
        else { print_stderr("Error during CFE: %s.", status_code_to_string(status)); } // Includes non-critical ERR_CALCULATION

        if (status < FAILURE) goto cleanup_search; // Abort on critical CFE error
        else { status = FAILURE; goto cleanup_search; } // Treat non-critical as overall failure to find solution
    }
    if (conv_count == 0) {
        // This case should have been mapped to ERR_CALCULATION inside CFE, but double-check.
        print_stderr("CFE generated no convergents within Q <= %lld.", max_overall_denominator);
        status = FAILURE; // No convergents means no solution possible via this method
        goto cleanup_search;
    }


    // --- Step 1.5: Find Simplest Exact Ratio (Ls/Ms) from CFE results ---
    // This ratio Ls/Ms will be used for the common factor (g/h) search later.
    for (int i = 0; i < conv_count; ++i) {
        if (convergents[i].q > 0 && convergents[i].p > 0) { // Check valid fraction
            double achieved_ratio = (double)convergents[i].p / (double)convergents[i].q;
            // Check relative difference carefully, avoid division by zero if target_ratio is tiny
            double rel_diff_exact = (fabs(target_ratio) > CFE_DBL_EPSILON)
                                    ? fabs((achieved_ratio - target_ratio) / target_ratio)
                                    : fabs(achieved_ratio - target_ratio);
            if (rel_diff_exact < EXACT_RATIO_EPSILON) {
                simplest_exact_found = true;
                L_simplest = convergents[i].p;
                M_simplest = convergents[i].q;
                break; // Found the first (smallest denominator) exact match
            }
        }
    }

    // --- Step 2: Process CFE Convergents (Direct Factor Permutations) ---
    // Iterate through each convergent L/M found by CFE.
    dots_printed_cfe = 0; // Reset dot counter before loop
    progress_msg_cfe_printed = false; // Reset message flag

    for (int i = 0; i < conv_count; ++i) {
        if (g_shutdown_requested) { status = FAILURE; break; }
        // Optimization: In Default+Balanced mode, if a dual match was already found, we can stop.
        if (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted &&
            inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default) {
            break;
        }

        // --- Print Progress Message/Dots ---
        // Only print progress for the potentially long Minimal Sum mode here
        if (search_mode == SEARCH_MINIMAL_SUM) {
            // Print header message only once, on the first relevant iteration (i>0)
            if (!progress_msg_cfe_printed && i > 0) {
                 printf("Progress (checking CFE convergents): ");
                 fflush(stdout);
                 progress_msg_cfe_printed = true;
            }
            // Print dots periodically after header is shown
            if (progress_msg_cfe_printed && i > 0 && i % DOT_FREQUENCY_CFE == 0) {
                 printf("."); fflush(stdout);
                 dots_printed_cfe++;
            }
        }
        // --- End Progress Message/Dots ---

        long long L = convergents[i].p;
        long long M = convergents[i].q;
        if (L <= 0 || M <= 0) continue; // Skip invalid convergents

        // Factor L = l1 * l2 for this convergent
        free(l_factors); l_factors = NULL; l_factor_count = 0; // Clear previous factors
        status = find_integer_factor_pairs(L, &l_factors, &l_factor_count);
        if (status != SUCCESS) {
            if (status < FAILURE) { print_stderr("Critical error factoring L=%lld: %s. Aborting.", L, status_code_to_string(status)); goto cleanup_search; }
            else { fprintf(stderr, "\nWarning: Skipping convergent L=%lld due to factor error %s.\n", L, status_code_to_string(status)); status = SUCCESS; continue; } // Warn and skip this convergent
        }
        if (l_factor_count == 0 && L > 1) { fprintf(stderr, "\nWarning: No factors found for L=%lld, skipping.\n", L); continue; } // Should not happen

        // Factor M = m1 * m2 for this convergent
        free(m_factors); m_factors = NULL; m_factor_count = 0; // Clear previous factors
        status = find_integer_factor_pairs(M, &m_factors, &m_factor_count);
        if (status != SUCCESS) {
            free(l_factors); l_factors = NULL; // Free L factors before handling M error
            if (status < FAILURE) { print_stderr("Critical error factoring M=%lld: %s. Aborting.", M, status_code_to_string(status)); goto cleanup_search; }
            else { fprintf(stderr, "\nWarning: Skipping convergent M=%lld due to factor error %s.\n", M, status_code_to_string(status)); status = SUCCESS; continue; } // Warn and skip
        }
        if (m_factor_count == 0 && M > 1) { fprintf(stderr, "\nWarning: No factors found for M=%lld, skipping.\n", M); free(l_factors); l_factors = NULL; continue; } // Should not happen

        // Check permutations L=l1*l2, M=m1*m2 -> Stages (l1,m1), (l2,m2) etc.
        for (int l_idx = 0; l_idx < l_factor_count; ++l_idx) {
             if (g_shutdown_requested || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
             long long li1 = l_factors[l_idx].factor1;
             long long li2 = l_factors[l_idx].factor2;

             for (int m_idx = 0; m_idx < m_factor_count; ++m_idx) {
                 if (g_shutdown_requested || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
                 long long mi1 = m_factors[m_idx].factor1;
                 long long mi2 = m_factors[m_idx].factor2;

                 // Evaluate candidate (l1, m1, l2, m2) and update best result if better
                 assess_candidate_and_update(li1, mi1, li2, mi2, max_lm_per_stage, inputs, search_mode, final_result, &first_dual_match_found_in_default);

                 // Evaluate candidate (l1, m2, l2, m1) if M factors are different
                 if (mi1 != mi2) {
                     assess_candidate_and_update(li1, mi2, li2, mi1, max_lm_per_stage, inputs, search_mode, final_result, &first_dual_match_found_in_default);
                 }

                 // Repeat with L factors swapped if they are different
                 if (li1 != li2) {
                     assess_candidate_and_update(li2, mi1, li1, mi2, max_lm_per_stage, inputs, search_mode, final_result, &first_dual_match_found_in_default);
                     if (mi1 != mi2) {
                         assess_candidate_and_update(li2, mi2, li1, mi1, max_lm_per_stage, inputs, search_mode, final_result, &first_dual_match_found_in_default);
                     }
                 }
             } // End m_idx loop
             if (g_shutdown_requested || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
        } // End l_idx loop

        // Free factors for *this* convergent before processing the next one
        free(l_factors); l_factors = NULL; l_factor_count = 0;
        free(m_factors); m_factors = NULL; m_factor_count = 0;

        if (g_shutdown_requested || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
    } // End CFE convergent loop

    // Finalize CFE progress message (if needed)
    if (status == SUCCESS && !g_shutdown_requested &&
        search_mode == SEARCH_MINIMAL_SUM &&
        conv_count > 0 && progress_msg_cfe_printed) // Check if header was printed
    {
        if (dots_printed_cfe == 0) printf("."); // Print confirmation dot if none printed
        printf("\n"); // Newline after CFE dots/message
        fflush(stdout);
    }

    // Free the main convergents array now we are done with it
    free(convergents); convergents = NULL; conv_count = 0;

    // Check status after CFE processing loop
    if (status != SUCCESS && status < FAILURE) goto cleanup_search; // Critical error during loop
    if (g_shutdown_requested) { status = FAILURE; goto cleanup_search; }


    // --- Step 3: Process Multiples of Simplest Exact Ratio (g/h Common Factors) ---
    // This explores solutions of the form L = Ls * g * h, M = Ms * g * h, rearranged into stages.
    // L1=g*ls1, M1=h*ms1, L2=h*ls2, M2=g*ms2  (where Ls=ls1*ls2, Ms=ms1*ms2)
    // Only run this if an exact simplest ratio Ls/Ms was found by CFE, and we haven't
    // already found a satisfactory dual match (if that optimization is active).
    if (simplest_exact_found && L_simplest > 0 && M_simplest > 0 &&
        !(search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted &&
          inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default))
    {
        // Factor the simplest L and M (reuse factor array variables)
        free(l_factors); l_factors = NULL; l_factor_count = 0;
        status = find_integer_factor_pairs(L_simplest, &l_factors, &l_factor_count);
        if (status != SUCCESS) {
            if (status < FAILURE) { print_stderr("Critical error factoring L_simplest=%lld: %s. Aborting.", L_simplest, status_code_to_string(status)); goto cleanup_search; }
            else { fprintf(stderr, "\nWarning: Cannot factor L_simplest=%lld (%s), skipping g/h search.\n", L_simplest, status_code_to_string(status)); status = SUCCESS; /* Skip g/h */ }
        } else if (l_factor_count == 0 && L_simplest > 1) { fprintf(stderr, "\nWarning: No factors found for L_simplest=%lld, skipping g/h search.\n", L_simplest); status = SUCCESS; }
        else { // L_simplest factored OK (or Ls=1)

            free(m_factors); m_factors = NULL; m_factor_count = 0;
            status = find_integer_factor_pairs(M_simplest, &m_factors, &m_factor_count);
            if (status != SUCCESS) {
                 free(l_factors); l_factors = NULL; // Free L factors
                 if (status < FAILURE) { print_stderr("Critical error factoring M_simplest=%lld: %s. Aborting.", M_simplest, status_code_to_string(status)); goto cleanup_search; }
                 else { fprintf(stderr, "\nWarning: Cannot factor M_simplest=%lld (%s), skipping g/h search.\n", M_simplest, status_code_to_string(status)); status = SUCCESS; /* Skip g/h */ }
            } else if (m_factor_count == 0 && M_simplest > 1) { fprintf(stderr, "\nWarning: No factors found for M_simplest=%lld, skipping g/h search.\n", M_simplest); free(l_factors); l_factors = NULL; status = SUCCESS; }
            else { // Both L_simplest and M_simplest factored successfully (or are 1).

                bool gh_minimal_processed = false; // Flag for Minimal Sum g/h optimization
                dots_printed_gh = 0;              // Reset g/h dot counter
                gh_iterations = 0;                // Reset g/h iteration counter
                progress_msg_gh_printed = false;  // Reset g/h progress message flag

                // Iterate through all factor pairs of L_simplest (ls1, ls2)
                for (int l_idx = 0; l_idx < l_factor_count; ++l_idx) {
                    if (g_shutdown_requested || (search_mode == SEARCH_MINIMAL_SUM && gh_minimal_processed) || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
                    long long ls1 = l_factors[l_idx].factor1;
                    long long ls2 = l_factors[l_idx].factor2;
                    if (ls1 <= 0 || ls2 <= 0) continue; // Factors must be positive

                    // Iterate through all factor pairs of M_simplest (ms1, ms2)
                    for (int m_idx = 0; m_idx < m_factor_count; ++m_idx) {
                         if (g_shutdown_requested || (search_mode == SEARCH_MINIMAL_SUM && gh_minimal_processed) || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
                         long long ms1 = m_factors[m_idx].factor1;
                         long long ms2 = m_factors[m_idx].factor2;
                         if (ms1 <= 0 || ms2 <= 0) continue; // Factors must be positive

                         // Determine maximum possible values for common factors g and h
                         // We need: g*ls1 <= max, g*ms2 <= max, h*ms1 <= max, h*ls2 <= max
                         long long common_g_max = LLONG_MAX;
                         long long common_h_max = LLONG_MAX;

                         // Calculate max g based on L1 = g*ls1 and M2 = g*ms2
                         if (max_lm_per_stage > 0) {
                             long long g_lim1 = (ls1 > 0) ? max_lm_per_stage / ls1 : LLONG_MAX;
                             long long g_lim2 = (ms2 > 0) ? max_lm_per_stage / ms2 : LLONG_MAX;
                             common_g_max = (g_lim1 < g_lim2) ? g_lim1 : g_lim2;
                         }
                         if (common_g_max <= 0) common_g_max = 1; // Ensure at least g=1 is checked

                         // Calculate max h based on M1 = h*ms1 and L2 = h*ls2
                         if (max_lm_per_stage > 0) {
                            long long h_lim1 = (ms1 > 0) ? max_lm_per_stage / ms1 : LLONG_MAX;
                            long long h_lim2 = (ls2 > 0) ? max_lm_per_stage / ls2 : LLONG_MAX;
                            common_h_max = (h_lim1 < h_lim2) ? h_lim1 : h_lim2;
                         }
                         if (common_h_max <= 0) common_h_max = 1; // Ensure at least h=1 is checked


                         // Loop through possible common factors g and h
                         // Start from 1 because g/h must be positive.
                         for (long long common_g = 1; common_g <= common_g_max; ++common_g) {
                              if (g_shutdown_requested || (search_mode == SEARCH_MINIMAL_SUM && gh_minimal_processed) || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;

                              for (long long common_h = 1; common_h <= common_h_max; ++common_h) {
                                   if (g_shutdown_requested || (search_mode == SEARCH_MINIMAL_SUM && gh_minimal_processed) || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;

                                   // --- Print g/h Progress ---
                                   gh_iterations++;
                                   if (!progress_msg_gh_printed) {
                                       printf("Progress (checking g/h factors): ");
                                       fflush(stdout);
                                       progress_msg_gh_printed = true;
                                   }
                                   if (gh_iterations % DOT_FREQUENCY_GH == 0) {
                                       printf("."); fflush(stdout);
                                       dots_printed_gh++;
                                   }
                                   // --- End g/h Progress ---


                                   // Construct candidate L1, M1, L2, M2 safely, checking for overflow
                                   long long L1 = 0, M1 = 0, L2 = 0, M2 = 0;
                                   bool overflow = false;

                                   // Calculate L1 = common_g * ls1
                                   if (ls1 > 0 && common_g > 0 && common_g > LLONG_MAX / ls1) overflow = true;
                                   else L1 = common_g * ls1;

                                   // Calculate M1 = common_h * ms1
                                   if (!overflow && ms1 > 0 && common_h > 0 && common_h > LLONG_MAX / ms1) overflow = true;
                                   else if (!overflow) M1 = common_h * ms1;

                                   // Calculate L2 = common_h * ls2
                                   if (!overflow && ls2 > 0 && common_h > 0 && common_h > LLONG_MAX / ls2) overflow = true;
                                   else if (!overflow) L2 = common_h * ls2;

                                   // Calculate M2 = common_g * ms2
                                   if (!overflow && ms2 > 0 && common_g > 0 && common_g > LLONG_MAX / ms2) overflow = true;
                                   else if (!overflow) M2 = common_g * ms2;

                                   // Check if constructed values are valid (positive)
                                   if (!overflow && (L1 <= 0 || M1 <= 0 || L2 <= 0 || M2 <= 0)) {
                                       overflow = true; // Treat non-positive results as unusable
                                   }

                                   // If overflow occurred during calculation, skip this g/h pair and potentially break loops
                                   if (overflow) {
                                       // If overflow happened even for h=1, increasing h won't help, so break h loop.
                                       // If it happened for g=1,h=1, increasing g might still work if h=1 is okay,
                                       // but it's safer to break h loop if overflow detected.
                                       break; // Break inner h loop
                                   }

                                   // If construction succeeded, assess this permutation
                                   // (assess_candidate will also check L/M <= max_lm_per_stage)
                                   assess_candidate_and_update(L1, M1, L2, M2, max_lm_per_stage, inputs, search_mode, final_result, &first_dual_match_found_in_default);

                                   // Minimal Sum Optimization: Check if we can stop after g=1, h=1
                                   if (search_mode == SEARCH_MINIMAL_SUM && common_g == 1 && common_h == 1) {

                                       if (!inputs->intermediate_rate_targeted ||
                                           inputs->intermediate_priority != PRIORITIZE_STRICT_INTERMEDIATE)
                                       {
                                           gh_minimal_processed = true;
                                           break;
                                       }
                                       else if (final_result->result_found && final_result->intermediate_match_found)
                                       {
                                           gh_minimal_processed = true;
                                           break;
                                       }
                                       // Break inner h loop (and outer g loop will break on next check)
                                   }
                              } // End common_h loop
                              // Break g loop if minimal sum g=1,h=1 processed or shutdown/optimisation triggered
                              if (g_shutdown_requested || (search_mode == SEARCH_MINIMAL_SUM && gh_minimal_processed) || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
                         } // End common_g loop
                         // Break m_idx loop if minimal sum g=1,h=1 processed or shutdown/optimisation triggered
                         if (g_shutdown_requested || (search_mode == SEARCH_MINIMAL_SUM && gh_minimal_processed) || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
                    } // End m_idx loop (M_simplest factors)
                     // Break l_idx loop if minimal sum g=1,h=1 processed or shutdown/optimisation triggered
                     if (g_shutdown_requested || (search_mode == SEARCH_MINIMAL_SUM && gh_minimal_processed) || (search_mode == SEARCH_DEFAULT && inputs->intermediate_rate_targeted && inputs->intermediate_priority == PRIORITIZE_BALANCED_DUAL && first_dual_match_found_in_default)) break;
                } // End l_idx loop (L_simplest factors)


                // Finalize g/h progress message (if needed)
                if (status == SUCCESS && !g_shutdown_requested && progress_msg_gh_printed) {
                    if (dots_printed_gh == 0) printf("."); // Confirmation dot
                    printf("\n"); // Newline after g/h dots/message
                    fflush(stdout);
                }


                // Free factors used for L_simplest/M_simplest
                free(l_factors); l_factors = NULL; l_factor_count = 0;
                free(m_factors); m_factors = NULL; m_factor_count = 0;
            } // End else (both Ls and Ms factored successfully)
        } // End else (Ls factored successfully)
    } // End if (simplest_exact_found and optimization allows)

cleanup_search:
    // Final cleanup of any dynamically allocated memory
    free(convergents); // Free if not already NULL
    free(l_factors);   // Free if not already NULL
    free(m_factors);   // Free if not already NULL

    // Determine final status
    if (status < FAILURE && status != SUCCESS) { return status; } // Return critical error status
    if (g_shutdown_requested) { return FAILURE; }

    // Return SUCCESS only if a valid result was found and stored
    return (final_result->result_found) ? SUCCESS : FAILURE;
} // End perform_two_stage_search

// --- Top-Level Function ---

/**
 * @brief Orchestrates the two-stage resampler search and result printing.
 * @details Prints context, calls the core search logic, prints results.
 */
StatusCode prompt_and_run_two_resamplers(const UserInputs* inputs)
{
    if (g_shutdown_requested) return FAILURE;
    if (!inputs) { print_stderr("Internal Error (prompt_two): NULL inputs pointer provided."); return ERR_INVALID_INPUT; }

    StatusCode final_overall_status = FAILURE; // Assume failure until success
    StatusCode run_status;
    TwoResamplerResult search_result; // Struct to hold the best result found
    long long max_lm_used = inputs->max_lm_per_stage; // Effective limit for this run
    int conv_count_for_msg = -1; // For informational message
    const char* mode_string = (inputs->search_mode == SEARCH_MINIMAL_SUM) ? "Minimal Sum" : "Custom Limit";

    // --- Print Context ---
    // (Prompts unchanged)
    printf("\n--- Two Rational Resampler Calculation (%s Mode) ---\n", mode_string);
    printf("Max L/M per stage: %lld\n", max_lm_used);
    if (inputs->intermediate_rate_targeted) {
        printf("Target Intermediate Rate: %.2f Hz (Priority: %s)\n", inputs->target_intermediate_rate, priority_to_string(inputs->intermediate_priority));
    }
    printf("Constraint: M1 > L1 is %s\n", inputs->allow_m1_greater_than_l1 ? "Allowed" : "NOT Allowed");
    printf("Target: Input Rate = %.2f Hz, Output Rate = %.2f Hz\n", inputs->input_rate, inputs->output_rate);
    double target_ratio_disp = (inputs->input_rate != 0) ? inputs->output_rate / inputs->input_rate : NAN;
    if (isfinite(target_ratio_disp)) printf("Target Ratio (output/input) = %.10f\n", target_ratio_disp);
    else printf("Target Ratio (output/input) = Invalid\n");
    printf("Match Tolerance (relative error): Overall=%.1e, Intermediate=%.1e\n", MATCH_TOLERANCE, INTERMEDIATE_MATCH_TOLERANCE);
    printf("Arbitrary Suggestion Threshold (relative error): %.1e\n", ARBITRARY_SUGGEST_ERROR_THRESHOLD);

    // Calculate and print max overall denominator used for CFE
    long long max_overall_denominator_disp = LLONG_MAX;
     if (max_lm_used > 0) {
         unsigned long long max_lm_ull_disp = (unsigned long long)max_lm_used;
         const unsigned long long SQRT_LLONG_MAX_APPROX_DISP = 3037000499ULL;
         if (max_lm_ull_disp > SQRT_LLONG_MAX_APPROX_DISP) { max_overall_denominator_disp = LLONG_MAX; }
         else { unsigned long long squared_disp = max_lm_ull_disp * max_lm_ull_disp; max_overall_denominator_disp = (squared_disp <= (unsigned long long)LLONG_MAX) ? (long long)squared_disp : LLONG_MAX; }
     }
     if (max_overall_denominator_disp <=0 && max_overall_denominator_disp != LLONG_MAX) max_overall_denominator_disp = 1;
    printf("CFE searching with max overall denominator Q <= %lld\n", max_overall_denominator_disp);

    // Add blank line before starting search message
    printf("\n");
    fflush(stdout);

    // --- Execute the Search ---
    if (g_shutdown_requested) {
         final_overall_status = FAILURE;
    } else {
        // Display waiting message / progress setup (Rely on internal progress now)
        printf("Starting search calculation. Please wait...\n\n");
        fflush(stdout); // Show this immediately


        // Call the core search function
        run_status = perform_two_stage_search(inputs, max_lm_used, inputs->search_mode, &search_result, &conv_count_for_msg);

        // --- Process Search Results ---
        // Newline added internally after progress dots now.

        if (g_shutdown_requested) {
             // Search function likely returned FAILURE, but confirm message
             printf("Search interrupted by user.\n");
             final_overall_status = FAILURE;
        } else if (run_status < FAILURE && run_status != SUCCESS) {
             // Core search function failed critically
             printf("Search failed critically: %s\n", status_code_to_string(run_status));
             final_overall_status = run_status; // Propagate critical error
        } else {
            // Search completed (either found something or found nothing suitable)

            // Print summary message (Prompts unchanged)
             if (conv_count_for_msg >= 0) { // Check if CFE ran
                 printf("%d CFE convergents processed...\n", conv_count_for_msg);
             }
             // Check if g/h loop might have run (Simplest exact ratio must have been found)
             // We don't have the simplest_exact_found flag here, but can give a generic message
             // if a result was found, as the g/h loop is the main source of complexity beyond CFE.
             if (search_result.result_found) {
                 printf("Checked factor pairs");
                 // Only mention g/h if simplest_exact_found was true, which we infer might have happened
                 // if Ls/Ms were factored inside the search. This is heuristic.
                 // A better approach might be to return a flag from perform_two_stage_search.
                 // For now, assume it might have run if a result exists.
                 printf(" and possibly common factors (g/h)...\n");
             }
             printf("Done searching!\n");
             printf("\n"); // Blank line after summary, before results
             fflush(stdout);


            // Print results using the final print function
            char title_buffer[100];
            snprintf(title_buffer, sizeof(title_buffer), "%s Search Results:", mode_string);
            print_two_resamplers_result(title_buffer, inputs, &search_result);

            // Determine final status based ONLY on whether a valid result was found
            // run_status might be SUCCESS (if search ran) even if result_found is false
            final_overall_status = search_result.result_found ? SUCCESS : FAILURE;
        }
    }

    return final_overall_status;
} // End prompt_and_run_two_resamplers
