// src/utils.c

#include "utils.h" // Include the header file for this module first
#include <stdio.h>
#include <stdlib.h> // For llabs (used indirectly by max_long_long logic if needed)
#include <stdarg.h> // For va_list, va_start, va_end
#include <string.h> // For strerror (used by some callers of print_stderr indirectly)
#include <errno.h>  // For errno (used by some callers of print_stderr indirectly)

// --- Function Implementations ---

/**
 * @brief Prints a formatted message to stderr, adding blank lines before and after.
 */
void print_stderr(const char* format, ...) {
    // Add the blank line before the message
    fprintf(stderr, "\n");

    // Print the actual formatted message
    va_list args;
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    // Add the trailing newline
    fprintf(stderr, "\n");
    fflush(stderr); // Ensure error message is displayed immediately
}

/**
 * @brief Converts a StatusCode enum value to its string representation.
 */
const char* status_code_to_string(StatusCode code) {
    switch (code) {
        case SUCCESS: return "SUCCESS";
        case FAILURE: return "FAILURE"; // General failure/user cancel/shutdown/not found
        case ERR_MEMORY_ALLOC: return "Memory Allocation Error";
        case ERR_INVALID_INPUT: return "Invalid Input Error";
        case ERR_CALCULATION: return "Calculation Error";
        case ERR_FILE_IO: return "File I/O Error"; // Placeholder
        case ERR_OVERFLOW: return "Arithmetic Overflow Error";
        default: return "Unknown Error Code";
    }
}

/**
 * @brief Returns the maximum of two long long values.
 *        Handles negative inputs defensively.
 */
long long max_long_long(long long a, long long b) {
    // Primarily intended for positive L/M values, but handle general case
    if (a < 0 && b < 0) return -1; // Or perhaps return the less negative one? Define behavior clearly. Let's return larger magnitude if both negative.
    if (a < 0) return b;
    if (b < 0) return a;
    // Both non-negative
    return (a > b) ? a : b;
}
